# https://gitlab.ssec.wisc.edu/sips/OrbNavClient
# https://sips.ssec.wisc.edu/orbnav#/

import json
import fnmatch
import os
import datetime
import glob
import warnings
from functools import partial
import geopy.distance
import xarray as xr
import pandas as pd
import numpy as np
import hashlib
import json
from datetime import datetime as dt
from itertools import chain
from pywapor.general.logger import log
from pywapor.collect import accounts
from pywapor.enhancers.apply_enhancers import apply_enhancer
from pywapor.general.processing_functions import save_ds, open_ds, remove_ds, adjust_timelim_dtype
from pywapor.general.curvilinear import create_grid, regrid
from pywapor.collect.protocol.crawler import download_url, download_urls
from pywapor.general.logger import log, adjust_logger
from pywapor.enhancers.other import drop_empty_times

def regrid_VNP(workdir, latlim, lonlim, dx_dy = (0.0033, 0.0033)):

    # Reformat bounding-box.
    bb = [lonlim[0], latlim[0], lonlim[1], latlim[1]]

    # Search for VNP02 images in workdir.
    ncs02 = glob.glob(os.path.join(workdir, "VNP02IMG.A*.nc"))

    # Make inventory of complete scenes in workdir.
    scenes = dict()
    for nc02 in ncs02:
        dt_str = ".".join(os.path.split(nc02)[-1].split(".")[1:3])
        ncs03 = glob.glob(os.path.join(workdir, f"VNP03IMG.{dt_str}*.nc"))
        ncs_cloud = glob.glob(os.path.join(workdir, f"CLDMSK_L2_VIIRS_SNPP.{dt_str}*.nc"))
        if len(ncs03) != 1 and len(ncs_cloud) != 1:
            continue
        else:
            nc03 = ncs03[0]
            nc_cloud = ncs_cloud[0]
        dt_obj = dt.strptime(dt_str, "A%Y%j.%H%M")
        scenes[dt_obj] = (nc02, nc03, nc_cloud)

    # Create list to store xr.Datasets from a single scene.
    dss = list()

    log.info(f"--> Processing {len(scenes)} scenes.").add()

    # Loop over the scenes.
    for i, (dt_obj, (nc02, nc03, nc_cloud)) in enumerate(scenes.items()):

        # Define tile output path.
        fp = os.path.join(workdir, f"VNP_{dt_obj:%Y%j%H%M}.nc")

        if os.path.isfile(fp):
            dss.append(open_ds(fp))
            continue

        # Open the datasets.
        ds1 = xr.open_dataset(nc02, group = "observation_data", decode_cf = False)
        ds2 = xr.open_dataset(nc03, group = "geolocation_data", decode_cf = False, chunks = "auto")
        ds3 = xr.open_dataset(nc_cloud, group = "geophysical_data", decode_cf = False, chunks = "auto")

        # Create cloud mask. (0=cloudy, 1=probably cloudy, 2=probably clear, 3=confident clear, -1=no result)
        cmask = ds3.Integer_Cloud_Mask.interp({
                        "number_of_lines": np.linspace(0-0.25, (ds3.number_of_lines.size-1)+0.25, ds3.number_of_lines.size*2),
                        "number_of_pixels": np.linspace(0-0.25, (ds3.number_of_pixels.size-1)+0.25, ds3.number_of_pixels.size*2)
                        }, kwargs = {"fill_value": "extrapolate"}).drop_vars(["number_of_lines", "number_of_pixels"])

        # Convert DN to Brightness Temperature using the provided loopup table.
        bt_da = ds1.I05_brightness_temperature_lut.isel(number_of_LUT_values = ds1.I05)

        # Rename some things.
        ds = ds2[["latitude", "longitude"]].rename({"latitude": "y", "longitude": "x"})

        # Chunk and mask invalid pixels.
        ds["bt"] = bt_da.chunk("auto").where((bt_da >= bt_da.valid_min) & 
                                            (bt_da <= bt_da.valid_max) & 
                                            (bt_da != bt_da._FillValue) &
                                            (ds1.I05_quality_flags == 0) &
                                            (ds2.quality_flag == 0) &
                                            (cmask == 3)
        )

        # Move `x` and `y` from variables to coordinates
        ds = ds.set_coords(["x", "y"])

        # Create mask for selecting the bounding-box in the data.
        buffer = 0.2
        mask = ((ds.y >= bb[1] - buffer) &
                (ds.y <= bb[3] + buffer) & 
                (ds.x >= bb[0] - buffer) & 
                (ds.x <= bb[2] + buffer))

        # Apply the mask.
        ds = ds.where(mask, drop = True)

        # Stop with scene if no valid data is found inside bb.
        if ds.bt.count().values == 0:
            log.warning(f"--> ({i+1}/{len(scenes)}) Skipping scene `{os.path.split(fp)[-1]}`, no valid data found inside bounding-box.")
            continue

        # Save intermediate file.
        fp_temp = os.path.join(workdir, f"VNP_{dt_obj:%Y%j%H%M}_temp.nc")
        ds = save_ds(ds, fp_temp, label = f"({i+1}/{len(scenes)}) Creating intermediate file.")

        # Create rectolinear grid.
        grid_ds = create_grid(ds, dx_dy[0], dx_dy[1], bb = bb)

        # Regrid from curvilinear to rectolinear grid.
        out = regrid(grid_ds, ds, max_px_dist = 3)

        # Set some metadata.
        out = out.rio.write_crs(4326)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category = FutureWarning)
            out = out.rio.clip_box(*bb)
        out.bt.attrs = {k: v for k, v in ds.bt.attrs.items() if k in ["long_name", "units"]}

        # Add time dimension.
        out = out.expand_dims({"time": 1}).assign_coords({"time": [dt_obj]})

        # Save regridded tile.
        out = save_ds(out, fp, encoding = "initiate", label = f"({i+1}/{len(scenes)}) Regridding `VNP_{dt_obj:%Y%j%H%M}.nc` to rectolinear grid.")

        dss.append(out)

        # Remove intermediate file.
        remove_ds(ds)

    log.sub()

    ds = xr.merge(dss)

    return ds

def boxtimes(latlim, lonlim, timelim, folder):
    """Check in which 6-min periods SUOMI NPPs swatch crosses a AOI
    defined by `latlim` and `lonlim` and remove night recordings.

    Parameters
    ----------
    latlim : list
        Latitude limits.
    lonlim : list
        Longitude limits.
    timelim : list
        Time limits.

    Returns
    -------
    list
        List containing year, doy and time at which SUOMI NPP passes over
        the bb during the day (night is filtered out).
    """

    # Convert to datetime object if necessary
    timelim = adjust_timelim_dtype(timelim)

    # Expand bb with 1500km (is half of SUOMI NPP swath width).
    ur = (latlim[1], lonlim[1])
    ll = (latlim[0], lonlim[0])
    very_ur = geopy.distance.distance(kilometers=1500).destination(ur, bearing=45)
    very_ll = geopy.distance.distance(kilometers=1500).destination(ll, bearing=45+180)

    # Define search kwargs.
    kwargs = {
        "sat" : '37849',
        "start" : timelim[0].strftime("%Y-%m-%dT%H:%M:%SZ"),
        "end" : timelim[1].strftime("%Y-%m-%dT%H:%M:%SZ"),
        "ur" : f"{int(np.ceil(very_ur.latitude))},{int(np.ceil(very_ur.longitude))}",
        "ll" : f"{int(np.floor(very_ll.latitude))},{int(np.floor(very_ll.longitude))}",
    }

    # Create a hash for the search kwargs to create a filename for this specific search, allowing
    # to reuse the file later if a similar search is made.
    search_hash = hashlib.sha1(json.dumps(kwargs, sort_keys=True).encode()).hexdigest()

    # Search boxtimes.
    base_url = "https://sips.ssec.wisc.edu/orbnav/api/v1/boxtimes.json"
    url = base_url + "?&" + "&".join([f"{k}={v}" for k, v in kwargs.items()])
    fp = os.path.join(folder, f"boxtimes_{search_hash}.json")
    _ = download_url(url, fp)
    with open(fp) as f:
        data = json.load(f)

    # Group to SUOMI NPP 6 minute tiles.
    dates = [dt.strptime(x[0], "%Y-%m-%dT%H:%M:%SZ") for x in chain.from_iterable(data["data"]) if x[3] > 40.]
    df = pd.DataFrame({"date": dates}).set_index("date")
    count = df.groupby(pd.Grouper(freq = "6min")).apply(lambda x: len(x))
    to_dl = count[count >= 1].index.values

    # Create list with relevant year/doy/times tuples.
    year_doy_time = [[pd.to_datetime(x).strftime("%Y"), pd.to_datetime(x).strftime("%j"), pd.to_datetime(x).strftime("%H%M")] for x in to_dl]

    return year_doy_time

def find_VIIRSL1_urls(year_doy_time, product, workdir, 
                        server_folders = {"VNP02IMG": 5110, "VNP03IMG": 5200, "CLDMSK_L2_VIIRS_SNPP": 5110}):
    """Given a list of year/doy/time tuples returns the exact urls for that
    datetime. Also check if the linked file already exists in workdir, in which
    case the url is omitted from the returned list.

    Parameters
    ----------
    year_doy_time : list
        List containing year, doy and time at which SUOMI NPP passes over
        the bb during the day (night is filtered out).
    product : str
        Name of the product to search.

    Returns
    -------
    list
        Urls linking to the product.
    """
    # Make empty url list.
    urls = list()

    # Loop over year/doy/times.
    for year, doy, t in year_doy_time:

        # Define url at which to find the json with the exact tile urls.
        url = f"https://ladsweb.modaps.eosdis.nasa.gov/archive/allData/{server_folders[product]}/{product}/{year}/{doy}.json"

        # Download the json.
        fp = os.path.join(workdir, f"{product}_{year}{doy}.json")
        if not os.path.isfile(fp):
            fp = download_url(url, fp)

        # Open the json and find the exact url of the required scene.
        all_scenes = [x["name"] for x in json.load(open(fp))]
        req_scenes = fnmatch.filter(all_scenes, f"{product}.A{year}{doy}.{t}.*.*.nc")

        # Check if a corrct url is found.
        if len(req_scenes) == 1:
            fn = req_scenes[0]
        else:
            continue

        # Check if the file already exists in workdir, otherwise add it to `urls`.
        url = f"https://ladsweb.modaps.eosdis.nasa.gov/archive/allData/{server_folders[product]}/{product}/{year}/{doy}/{fn}"
        fp = os.path.join(workdir, fn)
        if os.path.exists(fp):
            continue
        urls.append(url)

    return urls

def check_tiles(year_doy_time, latlim, lonlim, workdir, product = "VNP03IMG"):
    """Check if downloaded tiles actually contain data for the given AOI. If not
    removes its year/doy/time from the list.

    Parameters
    ----------
    year_doy_time : list
        List containing year, doy and time at which SUOMI NPP passes over
        the bb during the day (night is filtered out).
    latlim : list
        Latitude limits.
    lonlim : list
        Longitude limits.
    workdir : str
        Path to working directory
    product : str, optional
        name of the product to check, by default "VNP03IMG"

    Returns
    -------
    tuple
        Two lists, the first an updated `year_date_time`, the other contains the 
        values dropped from `year_date_time`.
    """
    # Create empty list to store dropped `year/doy/times`.
    dropped = list()

    # Reformat bounding-box.
    bb = [lonlim[0], latlim[0], lonlim[1], latlim[1]]

    # Loop over `year/doy/times`.
    for year, doy, time in year_doy_time:

        # Search for the relevant .nc file.
        fps = glob.glob(os.path.join(workdir, f"{product}.A{year}{doy}.{time}.*.nc"))
        assert len(fps) == 1
        fp = fps[0]

        # Open the geolocation data.
        ds = xr.open_dataset(fp, group = "geolocation_data", chunks = "auto")

        # Mask the irrelevant pixels.
        buffer = 0.2
        mask = ((ds.latitude >= bb[1] - buffer) &
                (ds.latitude <= bb[3] + buffer) & 
                (ds.longitude >= bb[0] - buffer) & 
                (ds.longitude <= bb[2] + buffer))

        # Check if there is relevant data in this tile.
        if mask.sum().values < 100:
            year_doy_time.remove([year, doy, time])
            dropped.append([year, doy, time])

    return year_doy_time, dropped

def check_geoloc_tile(fp, bb, min_pixels = 100):

    # Open the geolocation data.
    try: 
        ds = xr.open_dataset(fp, group = "geolocation_data", chunks = "auto")
    except OSError as e:
        if "HDF error" in str(e):
            log.info(f"--> {os.path.split(fp)[-1]} is corrupted, removing it.")
            os.remove(fp)
            return None
        else:
            raise e

    # Mask the irrelevant pixels.
    buffer = 0.2
    mask = ((ds.latitude >= bb[1] - buffer) &
            (ds.latitude <= bb[3] + buffer) & 
            (ds.longitude >= bb[0] - buffer) & 
            (ds.longitude <= bb[2] + buffer))

    # Check if there is relevant data in this tile.
    return mask.sum().values > min_pixels

def check_nc(fp, group = "observation_data"):

    try: 
        _ = xr.open_dataset(fp, group = group, chunks = "auto")
    except OSError as e:
        if "HDF error" in str(e):
            log.info(f"--> {os.path.split(fp)[-1]} is corrupted, removing it.")
            os.remove(fp)
            return None
        else:
            raise e

    return True

def default_vars(product_name, req_vars):

    variables = {
        "VNP02IMG": {"bt": {(), "bt"}},
    }

    req_dl_vars = {
        "VNP02IMG": {
            "bt": ["bt"],
        },
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def default_post_processors(product_name, req_vars = None):

    post_processors = {
        "VNP02IMG": {
            "bt": [partial(drop_empty_times, drop_vars = ["bt"])]
            },
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def download(folder, latlim, lonlim, timelim, product_name, req_vars,
                variables = None, post_processors = None):

    folder = os.path.join(folder, "VIIRSL1")
    if not os.path.exists(folder):
        os.makedirs(folder)

    appending = False
    fn = os.path.join(folder, f"{product_name}.nc")
    if os.path.isfile(fn):
        os.rename(fn, fn.replace(".nc", "_to_be_appended.nc"))
        existing_ds = open_ds(fn.replace(".nc", "_to_be_appended.nc"))
        if np.all([x in existing_ds.data_vars for x in req_vars]):
            existing_ds = existing_ds.close()
            os.rename(fn.replace(".nc", "_to_be_appended.nc"), fn)
            existing_ds = open_ds(fn)
            return existing_ds[req_vars]
        else:
            appending = True
            fn = os.path.join(folder, f"{product_name}_appendix.nc")
            req_vars = [x for x in req_vars if x not in existing_ds.data_vars]

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    # Reformat bounding-box.
    bb = [lonlim[0], latlim[0], lonlim[1], latlim[1]]

    # Find SUOMI NPP overpass times.
    year_doy_time = boxtimes(latlim, lonlim, timelim, folder)

    # Search for valid urls at overpass times for geolocation files.
    urls = find_VIIRSL1_urls(year_doy_time, "VNP03IMG", folder)

    # Create authentication header.
    token, _ = accounts.get('VIIRSL1')
    headers = {'Authorization': 'Bearer ' + token}

    # Try to download the geolocation data.
    checker_function = partial(check_geoloc_tile, bb = bb)
    relevant_fps = download_urls(urls, folder, headers = headers, checker_function = checker_function)

    # Check relevance of files that were already downloaded.
    fns = [os.path.split(x)[-1] for x in glob.glob(os.path.join(folder, "VNP03IMG*.nc"))]
    for fp in [os.path.join(folder, x) for x in fns if x not in [os.path.split(x)[-1] for x in urls]]:
        if check_geoloc_tile(fp, bb):
            relevant_fps.append(fp)

    # Convert the relevant netcdf names to year/doy/time entries.
    year_doy_time = [[os.path.split(x)[-1][{0:slice(10,14), 1:slice(14,17), 2:slice(18,22)}[y]] for y in range(3)] for x in relevant_fps]

    # Find urls of the cloudmask for the given year/doy/time.
    cloud_urls = find_VIIRSL1_urls(year_doy_time, "CLDMSK_L2_VIIRS_SNPP", folder)
    # Download the cloudmasks.
    _ =  download_urls(cloud_urls, folder, headers = headers, checker_function = partial(check_nc, group = "geophysical_data"))

    # Find urls of the actual data for the given year/doy/time.
    product_urls = find_VIIRSL1_urls(year_doy_time, product_name, folder)
    # Download the data product.
    _ =  download_urls(product_urls, folder, headers = headers, checker_function = check_nc)

    # Combine geolocations and data in rectilinear grid.
    ds = regrid_VNP(folder, latlim, lonlim, dx_dy = (0.0033, 0.0033))

    # Apply product specific functions.
    for var, funcs in post_processors.items():
        for func in funcs:
            ds, label = apply_enhancer(ds, var, func)
            log.info(label)

    ds_new = save_ds(ds, fn, encoding = "initiate", label = "Merging files.")

    if appending:
        ds = xr.merge([ds_new, existing_ds])
        lbl = f"Appending new variables (`{'`, `'.join(req_vars)}`) to existing file."
        ds = save_ds(ds, os.path.join(folder, f"{product_name}.nc"), encoding = "initiate", label = lbl)
        remove_ds(ds_new)
        remove_ds(existing_ds)
    else:
        ds = ds_new

    return ds

if __name__ == "__main__":

    folder = "/Users/hmcoerver/Local/20220325_20220415_test_data"
    adjust_logger(True, folder, "INFO")
    # latlim = [28.9, 29.7]
    # lonlim = [30.2, 31.2]
    # timelim = ["2022-06-01", "2022-06-02"]

    timelim = [datetime.date(2022, 3, 25), datetime.date(2022, 4, 15)]
    latlim = [29.4, 29.7]
    lonlim = [30.7, 31.0]

    product_name = "VNP02IMG"
    req_vars = ["bt"]

    variables = None
    post_processors = None

    ds = download(folder, latlim, lonlim, timelim, product_name, req_vars)


# -*- coding: utf-8 -*-
"""
This WAPOR toolbox is a set of functions to collect and run the WAPOR ET model
"""

from pywapor.collect import product, protocol, accounts

__all__ = ['product', 'protocol', 'accounts']



import os
from dask.diagnostics import ProgressBar
import numpy as np
from pywapor.general.logger import log
import xarray as xr
import numpy as np
import shutil
import glob
from datetime import datetime as dt
import warnings
import rasterio.warp
import pandas as pd
from pywapor.general.performance import performance_check

def log_example_ds(example_ds):
    if "source" in example_ds.encoding.keys():
        log.info(f"--> Using `{os.path.split(example_ds.encoding['source'])[-1]}` as reprojecting example.").add()
    else:
        log.info(f"--> Using variable `{list(example_ds.data_vars)[0]}` as reprojecting example.").add()
    shape = example_ds.y.size, example_ds.x.size
    res = example_ds.rio.resolution()
    log.info(f"> shape: {shape}, res: {abs(res[0]):.4f}° x {abs(res[1]):.4f}°.").sub()

def adjust_timelim_dtype(timelim):
    if isinstance(timelim[0], str):
        timelim[0] = dt.strptime(timelim[0], "%Y-%m-%d")
        timelim[1] = dt.strptime(timelim[1], "%Y-%m-%d")
    elif isinstance(timelim[0], np.datetime64):
        timelim[0] = dt.utcfromtimestamp(timelim[0].tolist()/1e9).date()
        timelim[1] = dt.utcfromtimestamp(timelim[1].tolist()/1e9).date()
    return timelim

def remove_ds(ds):
    if isinstance(ds, xr.Dataset):
        if "source" in ds.encoding.keys():
            fp = ds.encoding["source"]
            ds = xr.open_dataset(fp)
            ds = ds.close()
            os.remove(fp)
    elif isinstance(ds, str):
        if os.path.isfile(ds):
            fp = ds
            ds = xr.open_dataset(fp)
            ds = ds.close()
            os.remove(fp)        

def process_ds(ds, coords, variables, crs = None):

    ds = ds[list(variables.keys())]

    ds = ds.rename({v[0]:k for k,v in coords.items() if k in ["x", "y"]})
    ds = ds.rename({k: v[1] for k, v in variables.items()})

    if not isinstance(crs, type(None)):
        ds = ds.rio.write_crs(crs)

    ds = ds.rio.write_grid_mapping("spatial_ref")

    for var in [x for x in list(ds.variables) if x not in ds.coords]:
        if "grid_mapping" in ds[var].attrs.keys():
            del ds[var].attrs["grid_mapping"]

    ds = ds.sortby("y", ascending = False)
    ds = ds.sortby("x")

    ds.attrs = {}

    return ds

def make_example_ds(ds, folder, target_crs, bb = None):
    example_ds_fp = os.path.join(folder, "example_ds.nc")
    if os.path.isfile(example_ds_fp):
        example_ds = open_ds(example_ds_fp)
    else:
        if not isinstance(bb, type(None)):
            if ds.rio.crs != target_crs:
                loc_bb = transform_bb(target_crs, ds.rio.crs, bb)
            else:
                loc_bb = bb
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category = FutureWarning)
                ds = ds.rio.clip_box(*loc_bb)
            ds = ds.rio.pad_box(*loc_bb)
        ds = ds.rio.reproject(target_crs)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category = UserWarning)
            ds = ds.rio.clip_box(*bb)
        ds = ds.drop_vars(list(ds.data_vars))
        example_ds = save_ds(ds, example_ds_fp, encoding = "initiate", label = f"Creating example dataset.") # NOTE saving because otherwise rio.reproject bugs.
    return example_ds

@performance_check
def save_ds(ds, fp, decode_coords = "all", encoding = None, chunks = "auto", precision = 4):
    """Save a `xr.Dataset` as netcdf.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to save.
    fp : str
        Path to file to create.
    decode_coords : str, optional
        Controls which variables are set as coordinate variables when
        reopening the dataset, by default None.

    Returns
    -------
    xr.Dataset
        The newly created dataset.
    """
    temp_fp = fp.replace(".nc", "_temp.xx")

    folder = os.path.split(fp)[0]
    if not os.path.isdir(folder):
        os.makedirs(folder)

    if isinstance(chunks, dict):
        chunks = {dim: v for dim, v in chunks.items() if dim in ds.dims}

    ds = ds.chunk(chunks)

    if "y" in ds.coords:
        if len(ds.y.dims) == 1:
            ds = ds.sortby("y", ascending = False)
        ds = ds.rio.write_transform(ds.rio.transform(recalc=True))

    if encoding == "initiate":
        if not isinstance(precision, dict):
            precision = {var: precision for var in ds.data_vars}
        encoding = {var: {
                        "zlib": True,
                        "_FillValue": -9999,
                        "chunksizes": tuple([v[0] for _, v in ds[var].chunksizes.items()]),
                        "dtype": "int32", # determine_dtype(ds[var], -9999, precision.get(var)),
                        "scale_factor": 10**-precision.get(var, 0), 
                        } for var in ds.data_vars if np.all([spat in ds[var].coords for spat in ["x", "y"]])}
        for var in ds.data_vars:
            if "_FillValue" in ds[var].attrs.keys():
                _ = ds[var].attrs.pop("_FillValue")
        if "spatial_ref" in ds.coords:
            for var in ds.data_vars:
                if np.all([spat in ds[var].coords for spat in ["x", "y"]]):
                    ds[var].attrs.update({"grid_mapping": "spatial_ref"})
        for var in ds.coords:
            if var in ds.dims:
                encoding[var] = {"dtype": "float64"}

    with ProgressBar(minimum = 90, dt = 2.0):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="All-NaN slice encountered")
            ds.to_netcdf(temp_fp, encoding = encoding)

    ds = ds.close()

    os.rename(temp_fp, fp)

    ds = open_ds(fp, decode_coords = decode_coords, chunks = chunks)

    return ds

def open_ds(fp, decode_coords = "all", chunks = "auto", **kwargs):
    ds = xr.open_dataset(fp, decode_coords = decode_coords, chunks = chunks, **kwargs)
    return ds

def create_dummy_mask(x, y, sign = None, slope = None, xshift_fact = None, yshift_fact = None):
    if isinstance(sign, type(None)):
        sign = np.sign(np.random.random() - 0.5)
    if isinstance(slope, type(None)):
        slope = np.random.random()
    if isinstance(xshift_fact, type(None)):
        xshift_fact = np.random.random()
    if isinstance(yshift_fact, type(None)):
        yshift_fact = np.random.random()
    slope = sign*(slope * np.ptp(y) / np.ptp(x))
    yshift = {-1: y.max(), 1: y.min()}[np.sign(slope)]
    xshift = x.min()
    mask = (x - (xshift + xshift_fact * np.ptp(x))) * slope + (yshift + sign * yshift_fact * np.ptp(y))
    return y < mask

def create_dummy_ds(varis, fp = None, shape = (10, 1000, 1000), chunks = (-1, 500, 500), 
                    sdate = "2022-02-01", edate = "2022-02-11", precision = 2, min_max = [-1, 1],
                    latlim = [20,30], lonlim = [40, 50], data_generator = "random", mask_data = False):
    check = False
    if not isinstance(fp, type(None)):
        if os.path.isfile(fp):
            os.remove(fp)
    if not check:
        nt, ny, nx = shape
        dates = pd.date_range(sdate, edate, periods = nt)
        y,t,x = np.meshgrid(np.linspace(latlim[0], latlim[1], shape[1]), np.linspace(0, len(dates) + 1, len(dates)), np.linspace(lonlim[0],lonlim[1],shape[2]))
        if data_generator == "random":
            data = np.random.uniform(size = np.prod(shape), low = min_max[0], high=min_max[1]).reshape(shape)
        elif data_generator == "uniform":
            data = np.sqrt(x**2 + y**2)
            data = (data - data.min()) * ((min_max[1] - min_max[0]) / (data.max() - data.min())) + min_max[0]
        if mask_data:
            for i in range(nt):
                mask = create_dummy_mask(x[0,...], y[0,...])
                data[i,mask] = np.nan
        ds = xr.Dataset({k: (["time", "y", "x"], data) for k in varis}, coords = {"time": dates, "y": np.linspace(latlim[0], latlim[1], ny), "x": np.linspace(lonlim[0], lonlim[1], nx)})
        ds = ds.rio.write_crs(4326)
        if isinstance(chunks, tuple):
            chunks = {name: size for name, size in zip(["time", "y", "x"], chunks)}
        if not isinstance(fp, type(None)):
            ds = save_ds(ds, fp, chunks = chunks, encoding = "initiate", precision = precision, label = "Creating dummy dataset.")
    return ds

def determine_dtype(da, nodata, precision = None):
    if isinstance(precision, type(None)) and da.dtype.kind == "f":
        dtypes = [np.float16, np.float32, np.float64]
        precision = 0
        info = np.finfo
    else:
        dtypes = [np.int8, np.int16, np.int32, np.int64]
        info = np.iinfo
        if isinstance(precision, type(None)):
            precision = 0
    low = np.min([nodata, np.int0(np.floor(da.min().values * 10**precision))]) > [info(x).min for x in dtypes]
    high = np.max([nodata, np.int0(np.ceil(da.max().values * 10**precision))]) < [info(x).max for x in dtypes]
    check = np.all([low, high], axis = 0)
    if True in check:
        dtype = dtypes[np.argmax(check)]
    else:
        dtype = dtypes[-1]
        log.warning(f"--> Data for `{da.name}` with range [{np.int0(np.floor(da.min().values))}, {np.int0(np.ceil(da.max().values))}] doesnt fit inside dtype {dtype} with range [{np.iinfo(dtype).min}, {np.iinfo(dtype).max}] with a {precision} decimal precision.")
    return np.dtype(dtype).name

def create_wkt(latlim, lonlim):
    left = lonlim[0]
    bottom = latlim[0]
    right = lonlim[1]
    top = latlim[1]
    x = f"{left} {bottom},{right} {bottom},{right} {top},{right} {bottom},{left} {bottom}"
    return "GEOMETRYCOLLECTION(POLYGON((" + x + ")))"

def unpack(file, folder):
    fn = os.path.splitext(file)[0]
    shutil.unpack_archive(os.path.join(folder, file), folder)
    folder = [x for x in glob.glob(os.path.join(folder, fn + "*")) if os.path.isdir(x)][0]
    return folder

def transform_bb(src_crs, dst_crs, bb):
    bb =rasterio.warp.transform_bounds(src_crs, dst_crs, *bb, densify_pts=21)
    return bb

def calc_dlat_dlon(geo_out, size_X, size_Y, lat_lon = None):
    """
    Calculated the dimensions of each pixel in meter.

    Parameters
    ----------
    geo_out: list
        Geotransform function of the array.
    size_X: int
        Number of pixels in x-direction.
    size_Y: int
        Number of pixels in y-direction.

    Returns
    -------
    np.ndarray
        Size of every pixel in the y-direction in meters.
    dlon: array
        Size of every pixel in the x-direction in meters.
    """
    if isinstance(lat_lon, type(None)):
        # Create the lat/lon rasters
        lon = np.arange(size_X + 1)*geo_out[1]+geo_out[0] - 0.5 * geo_out[1]
        lat = np.arange(size_Y + 1)*geo_out[5]+geo_out[3] - 0.5 * geo_out[5]
    else:
        lat, lon = lat_lon

    dlat_2d = np.array([lat,]*int(np.size(lon,0))).transpose()
    dlon_2d =  np.array([lon,]*int(np.size(lat,0)))

    # Radius of the earth in meters
    R_earth = 6371000

    # Calculate the lat and lon in radians
    lonRad = dlon_2d * np.pi/180
    latRad = dlat_2d * np.pi/180

    # Calculate the difference in lat and lon
    lonRad_dif = abs(lonRad[:,1:] - lonRad[:,:-1])
    latRad_dif = abs(latRad[:-1] - latRad[1:])

    # Calculate the distance between the upper and lower pixel edge
    a = np.sin(latRad_dif[:,:-1]/2) * np.sin(latRad_dif[:,:-1]/2)
    clat = 2 * np.arctan2(np.sqrt(a), np.sqrt(1-a))
    dlat = R_earth * clat

    # Calculate the distance between the eastern and western pixel edge
    b = np.cos(latRad[1:,:-1]) * np.cos(latRad[:-1,:-1]) * np.sin(lonRad_dif[:-1,:]/2) * np.sin(lonRad_dif[:-1,:]/2)
    clon = 2 * np.arctan2(np.sqrt(b), np.sqrt(1-b))
    dlon = R_earth * clon

    return(dlat, dlon)

if __name__ == "__main__":

    folder = r"/Users/hmcoerver/Local/dummy_ds_test"

    varis = ["my_var"]
    shape = (10, 1000, 1000)
    sdate = "2022-02-02"
    edate = "2022-02-13" 
    fp = os.path.join(folder, "dummy_test.nc")
    precision = 2
    min_max = [-1, 1]

    ds = create_dummy_ds(varis, 
                    shape = shape, 
                    sdate = sdate, 
                    edate = edate, 
                    fp = fp,
                    precision = precision,
                    min_max = min_max,
                    )


import os
import urllib
from bs4 import BeautifulSoup
import warnings
import urllib
from joblib import Parallel, delayed
from functools import partial
import re
import numpy as np
import tqdm
from requests.exceptions import HTTPError
import time
import socket
import glob
import requests
from pywapor.general.logger import log, adjust_logger
from cachetools import cached, TTLCache
import ssl

@cached(cache=TTLCache(maxsize=2048, ttl=3600))
def find_paths(url, regex, node_type = "a", tag = "href", filter = None, session = None):
    if isinstance(session, type(None)):
        context = ssl._create_unverified_context()
        f = urllib.request.urlopen(url, context = context)
    else:
        file_object = session.get(url, stream = True)
        file_object.raise_for_status()
        f = file_object.content
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="It looks like you're parsing an XML document using an HTML parser.")
        soup = BeautifulSoup(f, "lxml")
    file_tags = soup.find_all(node_type, {tag: re.compile(regex)})
    if not isinstance(filter, type(None)):
        file_tags = np.unique([filter(tag) for tag in file_tags], axis = 0).tolist()
    return file_tags

def crawl(urls, regex, filter_regex, session, label_filter = None, list_out = False):

    def _filter(tag):
        url = tag["href"]
        date_str = re.findall(filter_regex, url)[0].replace("/", "")
        return (date_str, url)

    if list_out:
        crawled_urls = list()
    else:
        crawled_urls = dict()

    for url in tqdm.tqdm(urls.values(), delay = 20):
        raw_paths = find_paths(url, regex, filter = _filter, session = session)
        if list_out:
            crawled_urls += raw_paths
        else:
            crawled_urls = {**crawled_urls, **{dt: new_url for dt, new_url in raw_paths if dt in label_filter}}

    return crawled_urls

def _download_urls(urls, folder, session, fps = None, parallel = 0, headers = None,
                    max_tries = 10, wait_sec = 15):

    if isinstance(parallel, bool):
        parallel = {True: -1, False: 0}[parallel]

    if isinstance(fps, type(None)):
        fps = [os.path.join(folder, os.path.split(url)[-1]) for url in urls]
    else:
        assert len(fps) == len(urls)

    if parallel:
        backend = "loky"
        dler = partial(download_url, session = session, headers = headers, max_tries = max_tries, wait_sec = wait_sec)
        files = Parallel(n_jobs=parallel, backend = backend)(delayed(dler)(*x) for x in zip(urls, fps))
    else:
        files = list()
        for url, fp in zip(urls, fps):
            try:
                fp_out = download_url(url, fp, session = session, headers = headers, max_tries = max_tries, wait_sec = wait_sec)
                files.append(fp_out)
            except NameError as e:
                log.info(f"--> {e}")

    return files

def unzip(zipped):
    new, l = [], []
    for x in zipped:
        new.append(x[0])
        l.append(x[1])
    return new, l

def update_urls(urls, dled_files, relevant_fps, checker_function = None):
        
    if isinstance(urls, zip):
        urls, l = unzip(urls)
        fps_given = True
    else:
        l = [os.path.split(x)[-1] for x in urls]
        fps_given = False

    for fp in dled_files:

        if fp in relevant_fps:
            continue

        if isinstance(checker_function, type(None)):
            is_relevant = True
        else:
            is_relevant = checker_function(fp)

        if fps_given:
            x = fp
        else:
            x = os.path.split(fp)[-1]
        url_idx = l.index(x) if x in l else None

        if os.path.isfile(fp):
            if not isinstance(url_idx, type(None)):
                _ = urls.pop(url_idx)
                _ = l.pop(url_idx)
            if is_relevant:
                relevant_fps.append(fp)

    if fps_given:
        urls = zip(urls, l)

    return urls, relevant_fps

def download_urls(urls, folder, session = None, fps = None, parallel = 0, 
                    headers = None, checker_function = None, 
                    max_tries = 10, wait_sec = 15, meta_max_tries = 2):
    
    relevant_fps = list()
    try_n = 0
    try_again = True

    while try_again:

        dled_files = _download_urls(urls, folder, session, fps = fps, parallel = False, headers = headers, max_tries = max_tries, wait_sec = wait_sec)

        if not isinstance(fps, type(None)):
            urls = zip(urls, fps)

        urls, relevant_fps = update_urls(urls, dled_files, relevant_fps, checker_function = checker_function)

        if not isinstance(fps, type(None)):
            urls, fps = unzip(urls)

        try_again = len(urls) > 0 and try_n < meta_max_tries-1
        if try_again:
            try_n += 1
            log.info(f"--> Retrying to download {len(urls)} remaining urls.")

    if len(urls) > 0:
        log.warning(f"--> Didn't succeed to download {len(urls)} files.")
    return relevant_fps


def _download_url(url, fp, session = None, waitbar = True, headers = None):

    if os.path.isfile(fp):
        return fp

    folder, fn = os.path.split(fp)
    if not os.path.exists(folder):
        os.makedirs(folder)

    ext = os.path.splitext(fp)[-1]
    temp_fp = fp.replace(ext, "_temp")

    if isinstance(session, type(None)):
        session = requests.Session()

    file_object = session.get(url, stream = True, headers = headers)
    file_object.raise_for_status()

    if "Content-Length" in file_object.headers.keys():
        tot_size = int(file_object.headers["Content-Length"])
    else:
        tot_size = None

    if waitbar:
        wb = tqdm.tqdm(unit='Bytes', unit_scale=True, leave = False, 
                        total = tot_size, desc = fn)

    with open(temp_fp, 'wb') as z:
        for data in file_object.iter_content(chunk_size=1024):
            size = z.write(data)
            if waitbar:
                wb.update(size)

    os.rename(temp_fp, fp)

    return fp

def download_url(url, fp, session = None, waitbar = True, headers = None, 
                    max_tries = 10, wait_sec = 15):
    try_n = 0
    succes = False
    while try_n <= max_tries-1 and not succes:
        if try_n > 0:
            waiter = int((try_n**1.2) * wait_sec)
            log.info(f"--> Trying to download {fp}, attempt {try_n+1} of {max_tries} in {waiter} seconds.")
            time.sleep(waiter)
        try:
            fp = _download_url(url, fp, session = session, waitbar = waitbar, headers = headers)
            succes = True
        except socket.timeout as e:
            log.info(f"--> Server connection timed out.")
        except HTTPError as e:
            log.info(f"--> Server error {e}.")
        else:
            ...
        finally:
            try_n += 1
    
    if not succes:
        raise NameError(f"Could not download {url} after {max_tries} attempts.")

    return fp

if __name__ == "__main__":

    ...



"""
Generates input data for `pywapor.et_look`.
"""
from pywapor.collect import downloader
from pywapor.general.logger import log, adjust_logger
from pywapor.general import compositer
import pywapor.general.levels as levels
import datetime
import numpy as np
import os
import pandas as pd
import xarray as xr
from functools import partial
import pywapor.general.pre_defaults as defaults
from pywapor.general.variables import fill_attrs
from pywapor.enhancers.temperature import lapse_rate as _lapse_rate

def rename_vars(ds, *args):
    varis = ["p", "ra", "t_air", "t_air_min", "t_air_max", "u", "vp",
            "u2m", "v2m", "qv", "p_air", "p_air_0", "wv", "t_dew"]
    present_vars = [x for x in varis if x in ds.variables]
    ds = ds.rename({k: k + "_24" for k in present_vars})
    return ds

def lapse_rate(ds, *args):
    present_vars = [x for x in ds.variables if "t_air" in x]
    for var in present_vars:
        ds = _lapse_rate(ds, var)
    return ds

def calc_doys(ds, *args, bins = None):
    bin_doys = [int(pd.Timestamp(x).strftime("%j")) for x in bins]
    doy = np.mean([bin_doys[:-1], bin_doys[1:]], axis=0, dtype = int)
    if "time_bins" in list(ds.variables):
        ds["doy"] = xr.DataArray(doy, coords = ds["time_bins"].coords).chunk("auto")
    return ds

def add_constants(ds, *args):
    ds = ds.assign(defaults.constants_defaults())
    return ds

def main(folder, latlim, lonlim, timelim, sources = "level_1", bin_length = "DEKAD", enhancers = [lapse_rate]):
    """Prepare input data for `et_look`.

    Parameters
    ----------
    folder : str
        Path to folder in which to store (intermediate) data.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    sources : str | dict
        Configuration for each variable and source.
    bin_length : int | "DEKAD"
        Composite length.
    enhancers : list, optional
        Functions to apply to the xr.Dataset before creating the final
        output, by default "default".
    diagnostics : dict, optional
        Dictionary with coordinates and point-labels for which graphs can be 
        created.
    example_source : tuple, optional
        Which source to use for spatial alignment, overrides product selected
        through sources, by default None.

    Returns
    -------
    xr.Dataset
        Dataset with data for `pywapor.et_look`.
    """
    if not os.path.exists(folder):
        os.makedirs(folder)

    _ = adjust_logger(True, folder, "INFO")

    t1 = datetime.datetime.now()
    log.info("> PRE_ET_LOOK").add()

    if isinstance(sources, str):
        sources = levels.pre_et_look_levels(sources, bin_length = bin_length)

    bins = compositer.time_bins(timelim, bin_length)

    general_enhancers = enhancers + [rename_vars, fill_attrs, partial(calc_doys, bins = bins), add_constants]

    dss, sources = downloader.collect_sources(folder, sources, latlim, lonlim, [bins[0], bins[-1]], return_fps=False)

    ds = compositer.main(dss, sources, folder, general_enhancers, bins)

    t2 = datetime.datetime.now()
    log.sub().info(f"< PRE_ET_LOOK ({str(t2 - t1)})")

    return ds

if __name__ == "__main__":

#     enhancers = "default"
#     diagnostics = None
#     example_source = None
#     bin_length = "DEKAD"
    enhancers = [lapse_rate]

#     import pywapor

#     folder = r"/Users/hmcoerver/pywapor_notebooks_b"
#     latlim = [28.9, 29.7]
#     lonlim = [30.2, 31.2]
#     timelim = ["2021-07-01", "2021-07-11"]
#     composite_length = "DEKAD"

#     et_look_version = "v2"
#     export_vars = "default"

#     level = "level_1"

#     et_look_sources = pywapor.general.levels.pre_et_look_levels(level)

#     et_look_sources = {k: v for k, v in et_look_sources.items() if k in ["ndvi", "z"]}

    # et_look_sources["ndvi"]["products"] = [
    #     {'source': 'MODIS',
    #         'product_name': 'MOD13Q1.061',
    #         'enhancers': 'default'},
    #     {'source': 'MODIS', 
    #         'product_name': 'MYD13Q1.061', 
    #         'enhancers': 'default'},
    #     {'source': 'PROBAV',
    #         'product_name': 'S5_TOC_100_m_C1',
    #         'enhancers': 'default',
    #         'is_example': True}
    # ]

    # et_look_sources["r0"]["products"] = [
    #     {'source': 'MODIS',
    #         'product_name': 'MCD43A3.061',
    #         'enhancers': 'default'},
    #     {'source': 'PROBAV',
    #         'product_name': 'S5_TOC_100_m_C1',
    #         'enhancers': 'default'}
    # ]

    # se_root_sources = pywapor.general.levels.pre_se_root_levels(level)
    # se_root_sources["ndvi"]["products"] = et_look_sources["ndvi"]["products"]

    # from functools import partial
    # et_look_sources["se_root"]["products"] = [
    #     {'source': partial(pywapor.se_root.se_root, sources = se_root_sources),
    #         'product_name': 'v2',
    #         'enhancers': 'default'},]

    # ds = pywapor.pre_et_look.main(folder, latlim, lonlim, timelim, 
    #                                 sources = et_look_sources,
    #                                 bin_length = composite_length)






import datetime
import os
import json
import pywapor
import numpy as np
from functools import partial
import pywapor.collect.accounts as accounts
from shapely.geometry.polygon import Polygon
from pywapor.general.logger import log
from shapely.geometry import shape
from pywapor.collect.protocol.projections import get_crss
import pywapor.collect.protocol.opendap as opendap
import xarray as xr
from pywapor.general.processing_functions import open_ds, remove_ds, save_ds
from pywapor.general import bitmasks
import pandas as pd
import warnings

def fn_func(product_name, tile):
    fn = f"{product_name}_h{tile[0]:02d}v{tile[1]:02d}.nc"
    return fn

def url_func(product_name, tile):
    url = f"https://opendap.cr.usgs.gov/opendap/hyrax/{product_name}/h{tile[0]:02d}v{tile[1]:02d}.ncml.nc4?"
    return url

def tiles_intersect(latlim, lonlim):
    with open(os.path.join(pywapor.collect.__path__[0], "product/MODIS_tiles.geojson")) as f:
        features = json.load(f)["features"]
    aoi = Polygon.from_bounds(lonlim[0], latlim[0], lonlim[1], latlim[1])
    tiles = list()
    for feature in features:
        shp = shape(feature["geometry"])
        tile = feature["properties"]["Name"]
        if shp.intersects(aoi):
            h, v = tile.split(" ")
            htile = int(h.split(":")[-1])
            vtile = int(v.split(":")[-1])
            tiles.append((htile, vtile))
    return tiles

def shortwave_r0(ds, *args):
    ds["r0"] = 0.3 * ds["white_r0"] + 0.7 * ds["black_r0"]
    ds = ds.drop_vars(["white_r0", "black_r0"])
    return ds

def expand_time_dim(ds, *args):

    groups = ds.groupby(ds.lst_hour, squeeze = True)

    def _expand_hour_dim(x):
        hour = np.timedelta64(int(np.nanmedian(x.lst_hour.values) * 3600), "s")
        x = x.assign_coords({"hour": hour}).expand_dims("hour")
        return x

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Slicing with an out-of-order index")

        ds_expand = groups.map(_expand_hour_dim)

        ds_expand = ds_expand.stack({"datetime": ("hour","time")})

        new_coords = [time + hour for time, hour in zip(ds_expand.time.values, ds_expand.hour.values)]
        
        try: # new versions of xarray require to drop all dimensions of a multi-index
            ds_expand = ds_expand.drop_vars(["datetime", "hour", "time"])
        except ValueError: # old versions throw an error when trying to drop sub-dimensions of a multiindex.
            ds_expand = ds_expand.drop_vars(["datetime"])
        
        ds_expand = ds_expand.assign_coords({"datetime": new_coords}).rename({"datetime": "time"}).sortby("time")
        ds_expand = ds_expand.drop_vars(["lst_hour"])
        ds_expand = ds_expand.transpose("time", "y", "x")
        ds_expand = ds_expand.dropna("time", how="all")

    return ds_expand

def mask_bitwise_qa(ds, var, masker = "lst_qa", 
                product_name = "MOD11A1.061", flags = ["good_qa"]):

    new_data = ds[var]

    flag_bits = bitmasks.MODIS_qa_translator(product_name)
    mask = bitmasks.get_mask(ds[masker].astype("uint8"), flags, flag_bits)
    new_data = ds[var].where(mask, np.nan)
    ds = ds.drop_vars([masker])

    ds[var] = new_data

    return ds

def mask_qa(ds, var, masker = ("ndvi_qa", 1.0)):

    new_data = ds[var].where(ds[masker[0]] != masker[1], np.nan)
    ds = ds.drop_vars(masker[0])

    ds[var] = new_data

    return ds

def default_vars(product_name, req_vars):

    variables = {

        "MOD13Q1.061": {
                    "_250m_16_days_NDVI": [("time", "YDim", "XDim"), "ndvi"],
                    "_250m_16_days_pixel_reliability": [("time", "YDim", "XDim"), "ndvi_qa"],
                    "MODIS_Grid_16DAY_250m_500m_VI_eos_cf_projection": [(), "spatial_ref"],
                        },
        "MYD13Q1.061": {
                    "_250m_16_days_NDVI": [("time", "YDim", "XDim"), "ndvi"],
                    "_250m_16_days_pixel_reliability": [("time", "YDim", "XDim"), "ndvi_qa"],
                    "MODIS_Grid_16DAY_250m_500m_VI_eos_cf_projection": [(), "spatial_ref"],
                        },
        "MOD11A1.061": {
                    "LST_Day_1km": [("time", "YDim", "XDim"), "lst"],
                    "Day_view_time": [("time", "YDim", "XDim"), "lst_hour"],
                    "QC_Day": [("time", "YDim", "XDim"), "lst_qa"],
                    "MODIS_Grid_Daily_1km_LST_eos_cf_projection": [(), "spatial_ref"],
                        },
        "MYD11A1.061": {
                    "LST_Day_1km": [("time", "YDim", "XDim"), "lst"],
                    "Day_view_time": [("time", "YDim", "XDim"), "lst_hour"],
                    "QC_Day": [("time", "YDim", "XDim"), "lst_qa"],
                    "MODIS_Grid_Daily_1km_LST_eos_cf_projection": [(), "spatial_ref"],
                        },
        "MCD43A3.061": {
                    "Albedo_WSA_shortwave": [("time", "YDim", "XDim"), "white_r0"],
                    "Albedo_BSA_shortwave": [("time", "YDim", "XDim"), "black_r0"],
                    "BRDF_Albedo_Band_Mandatory_Quality_shortwave": [("time", "YDim", "XDim"), "r0_qa"],
                    "MOD_Grid_BRDF_eos_cf_projection": [(), "spatial_ref"]
        }
    }

    req_dl_vars = {
        "MOD13Q1.061": {
            "ndvi": ["_250m_16_days_NDVI", "_250m_16_days_pixel_reliability", "MODIS_Grid_16DAY_250m_500m_VI_eos_cf_projection"],
        },
        "MYD13Q1.061": {
            "ndvi": ["_250m_16_days_NDVI", "_250m_16_days_pixel_reliability", "MODIS_Grid_16DAY_250m_500m_VI_eos_cf_projection"],
        },
        "MOD11A1.061": {
            "lst": ["LST_Day_1km", "Day_view_time", "QC_Day", "MODIS_Grid_Daily_1km_LST_eos_cf_projection"],
        },
        "MYD11A1.061": {
            "lst": ["LST_Day_1km", "Day_view_time", "QC_Day", "MODIS_Grid_Daily_1km_LST_eos_cf_projection"],
        },
        "MCD43A3.061": {
            "r0": ["Albedo_WSA_shortwave", "Albedo_BSA_shortwave", "BRDF_Albedo_Band_Mandatory_Quality_shortwave", "MOD_Grid_BRDF_eos_cf_projection"],
        },
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def default_post_processors(product_name, req_vars = None):

    post_processors = {
        "MOD13Q1.061": {
            "ndvi": [mask_qa]
            },
        "MYD13Q1.061": {
            "ndvi": [mask_qa]
            },
        "MOD11A1.061": {
            "lst": [mask_bitwise_qa, expand_time_dim]
            },
        "MYD11A1.061": {
            "lst": [mask_bitwise_qa, expand_time_dim]
            },
        "MCD43A3.061": {
            "r0": [
                    shortwave_r0, 
                    partial(mask_qa, masker = ("r0_qa", 1.)),
                    ]
            },
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def download(folder, latlim, lonlim, timelim, product_name, req_vars,
                variables = None, post_processors = None):
    folder = os.path.join(folder, "MODIS")

    appending = False
    fn = os.path.join(folder, f"{product_name}.nc")
    if os.path.isfile(fn):
        os.rename(fn, fn.replace(".nc", "_to_be_appended.nc"))
        existing_ds = open_ds(fn.replace(".nc", "_to_be_appended.nc"))
        if np.all([x in existing_ds.data_vars for x in req_vars]):
            existing_ds = existing_ds.close()
            os.rename(fn.replace(".nc", "_to_be_appended.nc"), fn)
            existing_ds = open_ds(fn)
            return existing_ds[req_vars]
        else:
            appending = True
            fn = os.path.join(folder, f"{product_name}_appendix.nc")
            req_vars = [x for x in req_vars if x not in existing_ds.data_vars]

    if product_name == "MOD13Q1.061" or product_name == "MYD13Q1.061":
        timedelta = np.timedelta64(8, "D")
        timelim[0] = timelim[0] - pd.Timedelta(timedelta)
    elif product_name == "MCD43A3.061":
        timedelta = np.timedelta64(12, "h")
        timelim[0] = timelim[0] - pd.Timedelta(timedelta)
    else:
        timedelta = None

    tiles = tiles_intersect(latlim, lonlim)
    coords = {"x": ["XDim", lonlim], "y": ["YDim", latlim], "t": ["time",timelim]}
    
    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    data_source_crs = get_crss("MODIS")
    parallel = False
    spatial_tiles = True
    un_pw = accounts.get("NASA")
    request_dims = True
    ds_new = opendap.download(fn, product_name, coords, 
                variables, post_processors, fn_func, url_func, un_pw = un_pw, 
                tiles = tiles, data_source_crs = data_source_crs, parallel = parallel, 
                spatial_tiles = spatial_tiles, request_dims = request_dims,
                timedelta = timedelta)

    if appending:
        ds = xr.merge([ds_new, existing_ds])
        lbl = f"Appending new variables (`{'`, `'.join(req_vars)}`) to existing file."
        ds = save_ds(ds, os.path.join(folder, f"{product_name}.nc"), encoding = "initiate", label = lbl)
        remove_ds(ds_new)
        remove_ds(existing_ds)
    else:
        ds = ds_new

    return ds

if __name__ == "__main__":

    products = [
        # ('MCD43A3.061', ["r0"]),
        ('MOD11A1.061', ["lst"]),
        # ('MYD11A1.061', ["lst"]),
        # ('MOD13Q1.061', ["ndvi"]),
        # ('MYD13Q1.061', ["ndvi"]),
    ]

    folder = r"/Users/hmcoerver/Downloads/pywapor_test"
    # latlim = [26.9, 33.7]
    # lonlim = [25.2, 37.2]
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    timelim = [datetime.date(2020, 7, 1), datetime.date(2020, 7, 11)]


    for product_name, req_vars in products:
        ds = download(folder, latlim, lonlim, timelim, product_name, req_vars)
        print(ds.rio.crs, ds.rio.grid_mapping)


# -*- coding: utf-8 -*-
"""
Authors: Laust Færch
Module: Collect/PROBAV

Description:
This module preprocesses Landsat(7+8) data downloaded externaly

"""
from pywapor.collect.product.Landsat import C2L2SP

__all__ = ["C2L2SP"]

__version__ = '0.1'



"""Functions to process Landsat 7 and 8 Collection-2 Level-2 scenes into
NDVI, ALBEDO and LST GeoTIFFs that can be ingested into pywapor.pre_et_look.
"""
import rasterio
import glob
import os
import tarfile
import json
import shutil
from pywapor.general.processing_functions import save_ds, make_example_ds, remove_ds
import numpy as np
import xarray as xr
from osgeo import gdal
from datetime import datetime
from pywapor.general.logger import log
from pywapor.general import bitmasks
from pywapor.collect.protocol.crawler import download_url

def main(folder, max_lst_uncertainty = 2.5, final_bb = None):
    """Processes Landsat 7 or 8 Collection-2 Level-2 tar-files into GeoTIFFs
    for NDVI, LST and ALBEDO.

    Parameters
    ----------
    folder : str
        Path to folder containing Landsat 7 or 8 Collection-2 Level-2 
        tar-files.
    final_bb : list
        left, bottom, right, top.

    Returns
    -------
    ndvi_files : list
        List of paths to the NDVI files created.
    albedo_files : list
        List of paths to the ALBEDO files created.
    lst_files : list
        List of paths to the LST files created.
    """

    # Look for tar-files in the input folder.
    all_files = glob.glob(os.path.join(folder, "*.tar"))

    # Filter tar-files down to Landsat Level-2 files.
    files = [os.path.split(file)[-1] for file in all_files if "L2SP" in file]

    dss = list()

    example_ds = None

    # Loop over the tar-files.
    for file in files:

        # Unpack tar-file.
        ls_folder = untar(file, folder)

        ## ALBEDO & NDVI ##
        to_open = ["blue", "red", "green", "nir"]
        
        # Open the data.
        data, mtl, proj, geot = open_data(ls_folder, to_open)

        # Make cloud/shadow/quality mask.
        valid = open_mask(ls_folder, mtl, proj, geot)

        # Mask pixels.
        data = data.where(valid)

        # Calculate the albedo.
        albedo = calc_albedo(data)

        # Calculate the NDVI.
        ndvi = calc_ndvi(data)

        ## LST ##
        # Define which bands are required.
        to_open = ["therm", "therm_qa"]

        # Open the data.
        data = open_data(ls_folder, to_open)[0]
        
        # Create mask for pixels with an uncertainty greater than x Kelvin.
        valid = data.sel(band = "therm_qa") <= max_lst_uncertainty * 100

        # Apply mask.
        data = data.where(valid)
        
        # Calculate the lst.
        lst = calc_lst(data)

        # Merge the three variables into one dataset.
        ds = xr.merge([lst, ndvi, albedo])

        # Add time dimension to data arrays.
        ds = ds.expand_dims({"time": 1})

        # Set the correct time.
        date_str = mtl["IMAGE_ATTRIBUTES"]["DATE_ACQUIRED"]
        time_str = mtl["IMAGE_ATTRIBUTES"]["SCENE_CENTER_TIME"]
        datetime_str = date_str + " " + time_str.replace("Z", "")
        ds = ds.assign_coords({"time":[np.datetime64(datetime_str)]})

        target_crs = rasterio.crs.CRS.from_epsg(4326)

        # Clip and pad to bounding-box
        if isinstance(example_ds, type(None)):
            example_ds = make_example_ds(ds, os.path.join(folder, os.path.splitext(file)[0]), target_crs, bb = final_bb)
        ds = ds.rio.reproject_match(example_ds)
        ds = ds.assign_coords({"x": example_ds.x, "y": example_ds.y})

        # Save to netcdf
        fp = os.path.join(folder, os.path.splitext(file)[0] + ".nc")
        ds = save_ds(ds, fp, label = f"Processing `{file}`.")

        dss.append(ds)

        # Remove the untarred files and folder.
        shutil.rmtree(ls_folder)

    ds = xr.merge(dss, combine_attrs = "drop")

    fp = os.path.join(folder, "LANDSAT.nc")

    if os.path.isfile(fp):
        os.remove(fp)

    encoding = {v: {"zlib": True, "dtype": "float32"} for v in list(ds.data_vars)}
    encoding["time"] = {"dtype": "float64"}

    ds = save_ds(ds, fp, encoding = encoding, label = "Merging files.")

    for x in dss:
        remove_ds(x)

    return ds

def open_mask(ls_folder, mtl, proj, geot):
    """Create a mask to remove clouds/shadows and oversaturated pixels
    from the scene.

    Parameters
    ----------
    ls_folder : str
        Path to folder containing the scene images.
    mtl : dict
        Metadata of the landsat scene.
    proj : str
        Projection of the images to which the mask will be applied.
    geot : tuple
        Geotransform of the images to which the mask will be applied.

    Returns
    -------
    xr.DataArray
        Mask to apply to surface reflectance bands. True indicates pixels that
        should be kept, False indicates pixels that should be removed.
    """

    # Find Landsat spacecraft number.
    ls_number = ls_number_from_mtl(mtl)

    # Define which general bands to open.
    to_open = ["pixel_qa", "radsat_qa"]

    # Look up the filepaths.
    fps = find_files(ls_folder, to_open)

    # Open the data.
    qa_data = xr.concat([open_as_xr(fp, name) for name, fp in fps.items()], "band")
    
    # Check if projection and geotransform is the same as data to which the 
    # mask will be applied.
    _, _ = check_projs_geots(fps.values(), ref_proj_geot = (proj, geot))
    
    # Open the bit numbers.
    pixel_qa_bits = bitmasks.get_pixel_qa_bits(2, ls_number, 2)
    
    # Choose which labels to mask (see keys of 'pixel_qa_bits' for options).
    if ls_number in [8, 9]:
        pixel_qa_flags = ["dilated_cloud", "cirrus", "cloud", "cloud_shadow", "snow"]
    elif ls_number in [4, 5, 7]:
        pixel_qa_flags = ["dilated_cloud", "cloud", "cloud_shadow", "snow"]
    else:
        raise ValueError

    # Load array into RAM.
    qa_array = qa_data.sel(band="pixel_qa").band_data.values.astype("uint16")
    
    # Create the first mask.
    mask1 = bitmasks.get_mask(qa_array, pixel_qa_flags, pixel_qa_bits)

    # Open the bit numbers
    radsat_qa_bits = bitmasks.get_radsat_qa_bits(2, ls_number, 2)
    
    # Choose which labels to mask (all in this case).
    radsat_qa_flags = list(radsat_qa_bits.keys())
    
    # Load array into RAM.
    qa_array = qa_data.sel(band="radsat_qa").band_data.values.astype("uint16")
    
    # Create the second mask.
    mask2 = bitmasks.get_mask(qa_array, radsat_qa_flags, radsat_qa_bits)
    
    # Combine the masks.
    mask = np.invert(np.any([mask1, mask2], axis = 0))
    
    # Put the mask inside a xr.DataArray.
    mask_xr = xr.DataArray(data = np.transpose(mask), 
                            coords = qa_data.drop_dims(["band"]).coords)

    return mask_xr

def open_data(ls_folder, to_open):
    """Opens data in a folder created by unpacking a Landsat Collection-2 
    Level-2 tar-file, applies scale and offset factors, and masks invalid 
    values.

    Parameters
    ----------
    ls_folder : str
        Path to folder containing the scene images.
    to_open : list
        List with general band-names that need to be opened, e.g. ["red", "nir].

    Returns
    -------
    data : xr.Dataset
        Xarray dataset containing all the requested bands.
    mtl : dict
        Metadata of the landsat scene.
    proj : str
        Projection of the images.
    geot : tuple
        Geotransform of the images
    """

    # Open MTL file.
    mtl = open_mtl(ls_folder)

    # Check Landsat number.
    ls_number = ls_number_from_mtl(mtl)

    # Find the files that need to be opened.
    fps = find_files(ls_folder, to_open)

    # Check if all files have the same proj and geot.
    proj, geot = check_projs_geots(fps.values())

    # Open all the files into an xarray.Dataset.
    data = xr.concat([open_as_xr(fp, name) for name, fp in fps.items()], "band")

    # Find valid data range for each file.
    valid_ranges = valid_range()[ls_number]

    # Mask data outside valid range.
    masked_data = mask_invalid(data, valid_ranges)

    # Find relevant scales and offset factors for each file.
    scales_xr, offsets_xr = find_scales_offsets(fps, mtl)
    
    # Apply scale and offset to data.
    scaled_data = masked_data * scales_xr + offsets_xr
    
    # Add metadata.
    scaled_data.attrs = {"ls_number": ls_number}

    return scaled_data, mtl, proj, geot

def ls_number_from_mtl(mtl):
    """Find which Landsat spacecraft this mtl refers to.

    Parameters
    ----------
    mtl : dict
        Metadata of the landsat scene.

    Returns
    -------
    int
        Either 7 or 8.
    """
    return int(mtl["IMAGE_ATTRIBUTES"]["SPACECRAFT_ID"].split("_")[-1])

def scale_offset_from_mtl(fp, mtl, type):
    """Look up the scale and offset factor for a given filepath.

    Parameters
    ----------
    fp : str
        Path to GeoTIFF of a Landsat band.
    mtl : dict
        Metadata of the landsat scene.
    type : str
        Type of data in the fp, either "TEMPERATURE" or "REFLECTANCE".

    Returns
    -------
    float
        Scale factor to be applied to the fp.
    float
        Offset factor to be applied to the fp.
    """
    bandname = os.path.splitext(fp)[0].split("_")[-1][1:]

    addon = {True: "_ST_B", False: "_"}[type == "TEMPERATURE"]

    try:
        scale = float(mtl[f'LEVEL2_SURFACE_{type}_PARAMETERS'][f'{type}_MULT_BAND{addon}{bandname}'])
        offset = float(mtl[f'LEVEL2_SURFACE_{type}_PARAMETERS'][f'{type}_ADD_BAND{addon}{bandname}'])
    except KeyError:
        # log.warning(f"No scale/offset found for {fp}, setting to 1 and 0.")
        scale = 1.0
        offset = 0.0
    return scale, offset

def find_scales_offsets(fps, mtl):
    """Find scale and offset factors for multiple files and store them in
    xarray.DataArrays.

    Parameters
    ----------
    fps : dict
        Keys are general band-names, values are paths to files.
    mtl : dict
        Metadata of the landsat scene.

    Returns
    -------
    xarray.DataArray
        Array with the scale factors per band.
    xarray.DataArray
        Array with the offset factors per band.
    """

    scales_offsets = dict()
    for k, v in fps.items():
        if "therm" in k:
            scales_offsets[k] = scale_offset_from_mtl(v, mtl, "TEMPERATURE")
        else:
            scales_offsets[k] = scale_offset_from_mtl(v, mtl, "REFLECTANCE")

    scales_xr = xr.DataArray(data = np.array(list(scales_offsets.values()))[:,0], 
                        coords = {"band": list(scales_offsets.keys())})
    offsets_xr = xr.DataArray(data = np.array(list(scales_offsets.values()))[:,1], 
                        coords = {"band": list(scales_offsets.keys())})
    
    return scales_xr, offsets_xr

def proj_geot(fp, ds = None):
    """Find the projection and geotransform of a GeoTIFF file.

    Parameters
    ----------
    fp : str
        Path to GeoTIFF of a Landsat band.
    ds : gdal.Dataset, optional
        Alternatively, can use gdal.Dataset instead of fp, by default None

    Returns
    -------
    str
        Projection of the images.
    tuple
        Geotransform of the images.
    """
    if isinstance(ds, type(None)):
        ds = gdal.Open(fp)
    proj = ds.GetProjection()
    geot = ds.GetGeoTransform()
    return proj, geot

def acquired_at(mtl):
    """Find on which date and time the given scene was aqcuired.

    Parameters
    ----------
    mtl : dict
        Metadata of the landsat scene.

    Returns
    -------
    datetime.datetime
        Aqcuisition date and time.
    """
    date_time_str = mtl["IMAGE_ATTRIBUTES"]["DATE_ACQUIRED"] + mtl["IMAGE_ATTRIBUTES"]["SCENE_CENTER_TIME"]
    date_time = datetime.strptime(date_time_str[:-2], "%Y-%m-%d%H:%M:%S.%f")
    return date_time

def create_fn(var, mtl):
    """Create a filename conform pywapor.pre_et_look input files.

    Parameters
    ----------
    var : str
        Name of the variable for which to create a filename.
    mtl : dict
        Metadata of the landsat scene.

    Returns
    -------
    str
        Filename.
    """
    ls_number = ls_number_from_mtl(mtl)
    date_time = acquired_at(mtl)
    unit = {"lst": "K", "ndvi": "-", "r0": "-"}
    fn = f"{var}_LS{ls_number:01d}_{unit[var]}_-_{date_time:%Y.%m.%d.%H.%M}.tif"
    return fn

def calc_ndvi(data):
    """Calculate the Normalized Difference Vegetation Index (NDVI).

    Parameters
    ----------
    data : xarray.Dataset
        Dataset that needs to contain at least the coordinates "red" and "nir"
        in a dimension called 'band'.

    Returns
    -------
    xarray.Dataset
        Dataset that has a variable called "ndvi".
    """

    red = data.band_data.sel(band = "red")
    nir = data.band_data.sel(band = "nir")

    ndvi = (nir - red) / (nir + red)

    return xr.Dataset({"ndvi": ndvi})

def calc_albedo(data, ls_number = None, weights = None):
    """Calculate the albedo as a weighted average of the surface reflections.

    Parameters
    ----------
    data : xarray.Dataset
        Dataset that needs to contain at least the variables "red", "green",
        "blue", "nir", "swir1" and "swir2".
    ls_number : int, optional
        Landsat spacecraft number, can also be taken from 
        data.attrs["ls_number"], by default None.
    weights : dict, optional
        Dictionary whose keys are the same as the bandnames in data
        and its values the weights to be applied, uses default weights when 
        is None, by default None.

    Returns
    -------
    xarray.Dataset
        Dataset that has a variable called "ndvi".
    """

    if isinstance(weights, type(None)):
        if isinstance(ls_number, type(None)):
            ls_number = data.attrs["ls_number"]
        weights = albedo_weight()[ls_number]

    offset_band = xr.ones_like(data.isel(band=0)).expand_dims("band").assign_coords({"band": ["offset"]})
    offset_band = offset_band.where(data.band_data.isel(band=0).notnull())
    all_data = xr.concat([data, offset_band], "band")

    weights_xr = xr.DataArray(data = list(weights.values()), 
                            coords = {"band": list(weights.keys())})

    albedo = (all_data * weights_xr).sum("band", skipna = False).rename({"band_data": "r0"})

    return albedo

def calc_lst(data):
    """Calculate the land surface temperature.

    Parameters
    ----------
    data : xarray.Dataset
        Dataset that needs to contain at least the coordinate "therm" in the
        'band' dimension.

    Returns
    -------
    xarray.Dataset
        Dataset that has a variable called "lst".
    """
    return data.sel(band="therm").rename({"band_data": "lst"}).drop_vars(["band"])

def open_mtl(ls_folder):
    """Opens the MTL file that is included in every Landsat tar-file.

    Parameters
    ----------
    ls_folder : str
        Path to folder containing the scene images.

    Returns
    -------
    dict
        Metadata of the landsat scene.
    """
    fp_mtl = glob.glob(os.path.join(ls_folder, "*MTL.json"))

    assert len(fp_mtl) == 1, f"Multiple MTL-files found with '{os.path.join(ls_folder, '*MTL.json')}'. {fp_mtl}"

    with open(fp_mtl[0], "r") as json_mtl:
        mtl = json.load(json_mtl)['LANDSAT_METADATA_FILE']

    assert mtl["PRODUCT_CONTENTS"]["PROCESSING_LEVEL"] == "L2SP"
    
    return mtl

def untar(file, folder):
    """Unpack a tar-file.

    Parameters
    ----------
    file : str
        Path to tar-file.
    folder : str
        Path to folder in which the tar-file will be extracted.

    Returns
    -------
    str
        Folder in which the tar-file was extracted.
    """
    target_folder = os.path.join(folder, os.path.splitext(file)[0])

    with tarfile.open(os.path.join(folder, file)) as tar:
        tar.extractall(target_folder)

    return target_folder

def find_file(folder, string, expected_amount = None):
    """Search for files in 'folder' containing 'string' in their filepath. Can
    use "*" as wildcards.

    Parameters
    ----------
    folder : str
        Path to folder to be searched.
    string : str
        String to search in filenames.
    expected_amount : int, optional
        Check if more or less files were found than exepected, by default None

    Returns
    -------
    list
        Paths to the files that contain 'string'.
    """
    fps = glob.glob(os.path.join(folder, string))
    if isinstance(expected_amount, int):
        assert len(fps) == expected_amount, f"{fps}"
    return fps

def find_files(ls_folder, to_open):
    """Given a folder with an unpacked Landsat scene, look for filepaths of 
    bands specified by 'to_open'.

    Parameters
    ----------
    ls_folder : str
        Path to folder with unpacked Landsat scene.
    to_open : list
        List of strings, e.g. ["red", "green"].

    Returns
    -------
    dict
        Keys are the strings from 'to_open', values are matched filepaths.
    """
    search_paths = search_path()
    mtl = open_mtl(ls_folder)
    ls_number = ls_number_from_mtl(mtl)
    fps = {k: find_file(ls_folder, v, 1)[0] for k, v in search_paths[ls_number].items() if k in to_open}
    return fps

def open_as_xr(fp, name):
    """Open a file as a xarray.Dataset and change the default band coordinate
    to 'name'.

    Parameters
    ----------
    fp : str
        Path to GeoTIFF.
    name : str
        Name to be given as coordinate in the 'band' dimension.

    Returns
    -------
    xarray.Dataset
        Dataset with data.
    """
    return xr.open_dataset(fp, chunks = "auto").assign_coords({"band": [name]})

def albedo_weight():
    """Default weights to calculate albedo.

    Returns
    -------
    dict
        Default weights to be applied to the reflectanes in order to calculate
        albedo, for Landsat 7 and 8 Collection-2 Level-2 scenes.
    """
    weights = {
                5: {
                    "blue": 0.116,
                    "green": 0.010,
                    "red": 0.364,
                    "nir": 0.360,
                    "offset": 0.032,
                    },
                7: {
                    "blue": 0.085,
                    "green": 0.057,
                    "red": 0.349,
                    "nir": 0.359,
                    "offset": 0.033,
                    },
                8: {
                    "blue": 0.079,
                    "green": 0.083,
                    "red": 0.334,
                    "nir": 0.360,
                    "offset": 0.031,
                    },
                9: {
                    "blue": 0.079,
                    "green": 0.083,
                    "red": 0.334,
                    "nir": 0.360,
                    "offset": 0.031,
                    },
            }
    return weights

def search_path():
    """Link between general band-names and Landsat specific ones.

    Returns
    -------
    dict
        Keys are general names (e.g. "red", "nir") and values are strings to
        be used with glob.glob.
    """
    search_paths = {7: {
                    "blue":     "*SR_B1.TIF",
                    "green":    "*SR_B2.TIF",
                    "red":      "*SR_B3.TIF",
                    "nir":      "*SR_B4.TIF",
                    "swir1":    "*SR_B5.TIF",
                    "therm":    "*ST_B6.TIF",
                    "swir2":    "*SR_B7.TIF",
                    "pixel_qa": "*QA_PIXEL.TIF",
                    "radsat_qa":"*QA_RADSAT.TIF",
                    "therm_qa": "*ST_QA.TIF",
                    },
                
                5: {
                    "blue":     "*SR_B1.TIF",
                    "green":    "*SR_B2.TIF",
                    "red":      "*SR_B3.TIF",
                    "nir":      "*SR_B4.TIF",
                    "swir1":    "*SR_B5.TIF",
                    "therm":    "*ST_B6.TIF",
                    "swir2":    "*SR_B7.TIF",
                    "pixel_qa": "*QA_PIXEL.TIF",
                    "radsat_qa":"*QA_RADSAT.TIF",
                    "therm_qa": "*ST_QA.TIF",
                    },

                8: {
                    "blue":     "*SR_B2.TIF",
                    "green":    "*SR_B3.TIF",
                    "red":      "*SR_B4.TIF",
                    "nir":      "*SR_B5.TIF",
                    "swir1":    "*SR_B6.TIF",
                    "swir2":    "*SR_B7.TIF",
                    "therm":    "*ST_B10.TIF",
                    "pixel_qa": "*QA_PIXEL.TIF",
                    "radsat_qa":"*QA_RADSAT.TIF",
                    "therm_qa": "*ST_QA.TIF",
                    },
                9: {
                    "blue":     "*SR_B2.TIF",
                    "green":    "*SR_B3.TIF",
                    "red":      "*SR_B4.TIF",
                    "nir":      "*SR_B5.TIF",
                    "swir1":    "*SR_B6.TIF",
                    "swir2":    "*SR_B7.TIF",
                    "therm":    "*ST_B10.TIF",
                    "pixel_qa": "*QA_PIXEL.TIF",
                    "radsat_qa":"*QA_RADSAT.TIF",
                    "therm_qa": "*ST_QA.TIF",
                    },
                }
    return search_paths

def valid_range():
    """Valid DN ranges for Landsat 7 and 8 Collection-2 Level-2 bands.

    Returns
    -------
    dict
        Keys are general names (e.g. "red", "nir") and values are tuples with
        valid DN ranges.

    Notes
    -----
    These values are not stored in the MTL-file unfortunately and come from 
    the product-guides:
    https://www.usgs.gov/media/files/landsat-4-7-collection-2-level-2-science-product-guide
    https://www.usgs.gov/media/files/landsat-8-collection-2-level-2-science-product-guide 

    """


    valid_sr_range = {
        7: {
            "blue":     (7273, 43636),
            "green":    (7273, 43636),
            "red":      (7273, 43636),
            "nir":      (7273, 43636),
            "swir1":    (7273, 43636),
            "swir2":    (7273, 43636), # Assuming, not in table...
            "therm":    (1, 65535),
            "therm_qa":  (0, 32767),
            },
        8: {
            "aer":      (7273, 43636),
            "blue":     (7273, 43636),
            "green":    (7273, 43636),
            "red":      (7273, 43636),
            "nir":      (7273, 43636),
            "swir1":    (1, 65535),
            "swir2":    (7273, 43636),
            "therm":    (1, 65535),
            "therm_qa": (0, 32767),
            },
        }
    valid_sr_range[4] = valid_sr_range[7]
    valid_sr_range[5] = valid_sr_range[7]
    valid_sr_range[9] = valid_sr_range[8]

    return valid_sr_range

def mask_invalid(ds, valid_range):
    """Mask data that is outside of a valid range.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to be clipped.
    valid_range : dict
        Keys are general band-names, values are tuples of floats with lower and
        upper boundary.

    Returns
    -------
    xarray.Dataset
        Datasat in which values that lie outside the valid range are 
        set to nan.
    """

    min_xr = xr.DataArray(data = np.array(list(valid_range.values()))[:,0], 
                        coords = {"band": list(valid_range.keys())})
    max_xr = xr.DataArray(data = np.array(list(valid_range.values()))[:,1], 
                        coords = {"band": list(valid_range.keys())})

    return ds.where((ds.band_data >= min_xr) & (ds.band_data <= max_xr))

def check_projs_geots(files, ref_proj_geot = None):
    """Check if different files all have the same projection and geotransform.

    Parameters
    ----------
    files : list
        Paths to GeoTIFFS to be checked.
    ref_proj_geot : tuple, optional
        Extra projection and geotransform to check the files against. Default 
        is None.

    Returns
    -------
    str
        The uniform projection across all the files or the first projection in 
        case the projections are not uniform among the files.
    tuple
        The uniform geotransform across all the files or the first geotransform in 
        case the projections are not uniform among the files.    
    """
    projs_geots = np.array([proj_geot(fp) for fp in files], dtype=object)
    consistent_projs = np.unique(projs_geots[:,0]).size == 1     
    consistent_geots = np.unique(projs_geots[:,1]).size == 1
    assert consistent_projs
    assert consistent_geots
    proj = projs_geots[:,0][0]
    geot = projs_geots[:,1][0]
    if not isinstance(ref_proj_geot, type(None)):
        assert proj == ref_proj_geot[0]
        assert geot == ref_proj_geot[1]
    return proj, geot

def dl_landsat_test(folder):
    out_file = os.path.join(folder, "LE07_L2SP_177040_20210707_20210802_02_T1.tar")
    url = 'https://storage.googleapis.com/fao-cog-data/LE07_L2SP_177040_20210707_20210802_02_T1.tar'
    download_url(url, out_file, None)
    return out_file

if __name__ == "__main__":

    max_lst_uncertainty = 1.0

    folder = r"/Users/hmcoerver/On My Mac/ndvi_r0_test"

    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]

    left = lonlim[0]
    bottom = latlim[0]
    right = lonlim[1]
    top = latlim[1]

    final_bb = [left, bottom, right, top]

    # ds = main(folder, max_lst_uncertainty = max_lst_uncertainty, final_bb = final_bb)

# #%%

# import xarray as xr
# from pywapor.general.processing_functions import save_ds, open_ds
# import rasterio


# latlim = [28.9, 29.7]
# lonlim = [30.2, 31.2]
# left = lonlim[0]
# bottom = latlim[0]
# right = lonlim[1]
# top = latlim[1]

# final_bb = [left, bottom, right, top]

# def transform_bb(src_crs, dst_crs, bb):
#     bb =rasterio.warp.transform_bounds(src_crs, dst_crs, *bb, densify_pts=21)
#     return bb

# ds = open_ds(r"/Users/hmcoerver/On My Mac/ndvi_r0_test/test_in.nc")

# target_crs = rasterio.crs.CRS.from_epsg(4326)

# bb = transform_bb(target_crs, ds.rio.crs, final_bb)

# ds = ds.rio.clip_box(*bb)
# ds = ds.rio.pad_box(*bb)

# ds = ds.rio.reproject(target_crs)




import os
import json
import pywapor
import getpass
import sys
import requests
import time
from pywapor.general.logger import log
from cryptography.fernet import Fernet
import cdsapi
from sentinelsat import SentinelAPI

def setup(account):
    """Asks, saves and tests a username/password combination for `account`.

    Parameters
    ----------
    account : {"NASA" | "VITO" | "WAPOR" | "ECMWF" | "SENTINEL" | "VIIRSL1"}
        Which un/pw combination to store.
    """

    folder = os.path.dirname(os.path.realpath(pywapor.__path__[0]))
    filename = "secret.txt"
    filehandle = os.path.join(folder, filename)
    json_filehandle = os.path.join(folder, "keys.json")

    if not os.path.exists(filehandle):
        create_key()
    
    f = open(filehandle,"r")
    key = f.read()
    f.close()
    
    cipher_suite = Fernet(key.encode('utf-8'))

    if os.path.exists(json_filehandle):
        # open json file  
        with open(json_filehandle) as f:
            datastore = f.read()
        obj = json.loads(datastore)      
        f.close()
    else:
        obj = {}

    if account == "WAPOR":
        API_Key = input(f"{account} API token: ")
        API_Key_crypt = cipher_suite.encrypt(("%s" %API_Key).encode('utf-8'))
        obj[account] = [str(API_Key_crypt.decode("utf-8"))]
    if account == "VIIRSL1":
        API_Key = input(f"{account} API token: ")
        API_Key_crypt = cipher_suite.encrypt(("%s" %API_Key).encode('utf-8'))
        obj[account] = [str(API_Key_crypt.decode("utf-8"))]
    elif account == "ECMWF":
        API_Key = input(f"{account} CDS API key: ")
        API_Key_crypt = cipher_suite.encrypt(("%s" %API_Key).encode('utf-8'))
        obj[account] = [str(API_Key_crypt.decode("utf-8"))]
    else:
        account_name = input(f"{account} username: ")
        pwd = getpass.getpass(f"{account} password: ")            
        account_name_crypt = cipher_suite.encrypt(("%s" %account_name).encode('utf-8'))
        pwd_crypt = cipher_suite.encrypt(("%s" %pwd).encode('utf-8'))
        obj[account] = ([str(account_name_crypt.decode("utf-8")), str(pwd_crypt.decode("utf-8"))])

    # save extent in task
    with open(json_filehandle, 'w') as outfile:
        json.dump(obj, outfile)        

    if account == "NASA":
        log.info("--> Testing NASA un/pw.")
        succes = nasa_account()
        if succes:
            log.info("--> NASA un/pw working.")
        else:
            _ = obj.pop("NASA")
            with open(json_filehandle, 'w') as outfile:
                json.dump(obj, outfile)
            sys.exit(f"Please fix your NASA account.")            

    if account == "VITO":
        log.info("--> Testing VITO un/pw.")
        succes = vito_account()
        if succes:
            log.info("--> VITO un/pw working.")
        else:
            _ = obj.pop("VITO")
            with open(json_filehandle, 'w') as outfile:
                json.dump(obj, outfile)
            sys.exit(f"Please fix your VITO account.")  

    if account == "WAPOR":
        log.info("--> Testing WAPOR token.")
        succes = wapor_account()
        if succes:
            log.info("--> WAPOR token working.")
        else:
            _ = obj.pop("WAPOR")
            with open(json_filehandle, 'w') as outfile:
                json.dump(obj, outfile)
            sys.exit(f"Please fix your WAPOR token.")

    if account == "VIIRSL1":
        log.info("--> Testing VIIRSL1 token.")
        succes = True
        if succes:
            log.info("--> VIIRSL1 token working.")
        else:
            _ = obj.pop("VIIRSL1")
            with open(json_filehandle, 'w') as outfile:
                json.dump(obj, outfile)
            sys.exit(f"Please fix your VIIRSL1 token.")  

    if account == "ECMWF":
        log.info("--> Testing ECMWF key.")
        succes = ecmwf_account()
        if succes:
            log.info("--> ECMWF key working.")
        else:
            _ = obj.pop("ECMWF")
            with open(json_filehandle, 'w') as outfile:
                json.dump(obj, outfile)
            sys.exit(f"Please fix your ECMWF key.")

    if account == "SENTINEL":
        log.info("--> Testing Sentinel key.")
        succes = sentinel_account()
        if succes:
            log.info("--> Sentinel key working.")
        else:
            _ = obj.pop("Sentinel")
            with open(json_filehandle, 'w') as outfile:
                json.dump(obj, outfile)
            sys.exit(f"Please fix your Sentinel key.") 

    return

def get(account):
    """Loads a required username/password.

    Parameters
    ----------
    account : {"NASA" | "VITO" | "WAPOR" | "ECMWF" | "SENTINEL" | "VIIRSL1"}
        Which un/pw combination to load.
    """

    folder = os.path.dirname(os.path.realpath(pywapor.__path__[0]))
    filename = "secret.txt"
    key_file = os.path.join(folder, filename)
    json_file = os.path.join(folder, "keys.json")
    
    if not os.path.exists(key_file):
        create_key()
        if os.path.exists(json_file):
            print("removing old/invalid json file")
            os.remove(json_file)
    
    if not os.path.exists(json_file):
        setup(account)

    f = open(key_file,"r")
    key = f.read()
    f.close()
    
    cipher_suite = Fernet(key.encode('utf-8'))    
    
    # open json file  
    with open(json_file) as f:
        datastore = f.read()
    obj = json.loads(datastore)      
    f.close()

    if account not in obj.keys():
        setup(account)
        obj = None 
        with open(json_file) as f:
            datastore = f.read()
        obj = json.loads(datastore)      
        f.close()       

    if account == "WAPOR":
        username_crypt = obj[account]
        username = cipher_suite.decrypt(username_crypt[0].encode("utf-8"))
        pwd = b''
    elif account == "VIIRSL1":
        username_crypt = obj[account]
        username = cipher_suite.decrypt(username_crypt[0].encode("utf-8"))
        pwd = b''
    elif account == "ECMWF":
        pwd_crypt = obj[account]
        pwd = cipher_suite.decrypt(pwd_crypt[0].encode("utf-8"))
        username = b'https://cds.climate.copernicus.eu/api/v2'
    else:
        username_crypt, pwd_crypt = obj[account]
        username = cipher_suite.decrypt(username_crypt.encode("utf-8"))
        pwd = cipher_suite.decrypt(pwd_crypt.encode("utf-8"))  
    
    return(str(username.decode("utf-8")), str(pwd.decode("utf-8")))

def create_key():
    """Generates a key file.
    """
    folder = os.path.dirname(os.path.realpath(pywapor.__path__[0]))
    filename = "secret.txt"
    filehandle = os.path.join(folder, filename)

    if os.path.exists(filehandle):
        os.remove(filehandle)
    
    with open(filehandle,"w+") as f:
        f.write(str(Fernet.generate_key().decode("utf-8")))

def vito_account(user_pw = None):
    """Check if the given or stored VITO username/password combination
    is correct. Accounts can be created on https://www.vito-eodata.be.

    Parameters
    ----------
    user_pw : tuple, optional
        ("username", "password") to check, if `None` will try to load the 
        password from the keychain, by default None.

    Returns
    -------
    bool
        True if the password works, otherwise False.
    """
    n_max = 3
    succes = False
    n = 1

    while not succes and n <= n_max:

        folder = os.path.dirname(os.path.realpath(pywapor.__path__[0]))
        test_url = r"https://www.vito-eodata.be/PDF/datapool/Free_Data/PROBA-V_300m/S1_TOC_-_300_m_C1/2014/1/15/PV_S1_TOC-20140115_333M_V101/PROBAV_S1_TOC_20140115_333M_V101.VRG"
        test_file = os.path.join(folder, "vito_test.vrg")

        if os.path.isfile(test_file):
            os.remove(test_file)

        if not isinstance(user_pw, type(None)):
            username, password = user_pw
        else:
            username, password = get('VITO')

        x = requests.get(test_url, auth = (username, password))

        if x.ok:

            with open(test_file, 'w+b') as z:
                z.write(x.content)

                statinfo = os.stat(test_file)
                succes = statinfo.st_size == 15392
                os.remove(test_file)
                if not succes:
                    error = "something went wrong."

        else:
            error = "wrong username/password."
            succes = False
            if os.path.isfile(test_file):
                os.remove(test_file)

        if not succes:
            log.warning(f"Try {n}/{n_max} failed.")
            time.sleep(3)
            
        n += 1

    if not succes:
        log.warning(error)

    return succes

def nasa_account(user_pw = None):
    """Check if the given or stored NASA username/password combination is 
    correct. Accounts can be created on https://urs.earthdata.nasa.gov/users/new.

    Parameters
    ----------
    user_pw : tuple, optional
        ("username", "password") to check, if `None` will try to load the 
        password from the keychain, by default None.

    Returns
    -------
    bool
        True if the password works, otherwise False.
    """
    n_max = 3
    succes = False
    n = 1

    while not succes and n <= n_max:

        folder = os.path.dirname(os.path.realpath(pywapor.__path__[0]))
        test_url = r"https://goldsmr4.gesdisc.eosdis.nasa.gov/data/MERRA2/M2T3NXGLC.5.12.4/1987/08/MERRA2_100.tavg3_2d_glc_Nx.19870801.nc4"
        test_file = os.path.join(folder, "nasa_test.nc4")

        if os.path.isfile(test_file):
            os.remove(test_file)

        if not isinstance(user_pw, type(None)):
            username, password = user_pw
        else:
            username, password = get('NASA')

        x = requests.get(test_url, allow_redirects = False)
        y = requests.get(x.headers['location'], auth = (username, password))

        if x.ok and y.ok:

            with open(test_file, 'w+b') as z:
                z.write(y.content)
        
            if os.path.isfile(test_file):
                statinfo = os.stat(test_file)
                succes = statinfo.st_size == 3963517
                os.remove(test_file)
                if not succes:
                    error = "please add 'NASA GESDISC DATA ARCHIVE' to 'Approved Applications'."
            else:
                succes = False

        else:
            error = "wrong username/password."
            succes = False
            if os.path.isfile(test_file):
                os.remove(test_file)

        if not succes:
            log.warning(f"Try {n}/{n_max} failed.")
            time.sleep(3)
            
        n += 1

    if not succes:
        log.warning(error)

    return succes

def wapor_account(user_pw = None):
    """Check if the given or stored WAPOR token is 
    correct. Accounts can be created on https://wapor.apps.fao.org/home/WAPOR_2/1.

    Parameters
    ----------
    user_pw : tuple, optional
        ("", "token") to check, if `None` will try to load the 
        password from the keychain, by default None.

    Returns
    -------
    bool
        True if the password works, otherwise False.
    """
    n_max = 3
    succes = False
    n = 1

    while not succes and n <= n_max:

        if not isinstance(user_pw, type(None)):
            username, _ = user_pw
        else:
            username, _ = get('WAPOR')

        sign_in= 'https://io.apps.fao.org/gismgr/api/v1/iam/sign-in'
        resp_vp=requests.post(sign_in,headers={'X-GISMGR-API-KEY':username})
        
        if resp_vp.ok:
            succes = True
        else:
            error = "wrong token."
            succes = False

        if not succes:
            log.warning(f"Try {n}/{n_max} failed.")
            time.sleep(3)
            
        n += 1

    if not succes:
        log.warning(error)

    return succes

def viirsl1_account(user_pw = None):
    """Check if the given or stored VIIRSL1 token is 
    correct. Accounts can be created on https://ladsweb.modaps.eosdis.nasa.gov/.

    Parameters
    ----------
    user_pw : tuple, optional
        ("", "token") to check, if `None` will try to load the 
        password from the keychain, by default None.

    Returns
    -------
    bool
        True if the password works, otherwise False.
    """
    n_max = 3
    succes = False
    n = 1

    while not succes and n <= n_max:

        if not isinstance(user_pw, type(None)):
            username, _ = user_pw
        else:
            username, _ = get('VIIRSL1')

        headers = {'Authorization': 'Bearer ' + username}
        test_url = "https://ladsweb.modaps.eosdis.nasa.gov/archive/allData/5110/VNP02IMG/2018/008/VNP02IMG.A2018008.2348.001.2018061005026.nc"
        response = requests.Session().get(test_url, headers=headers, stream = True)
        for data in response.iter_content(chunk_size=1024):
            succes = b'DOCTYPE' not in data
            if not succes:
                error = "wrong token."
            break

        if not succes:
            log.warning(f"Try {n}/{n_max} failed.")
            time.sleep(3)
            
        n += 1

    if not succes:
        log.warning(error)

    return succes

def sentinel_account(user_pw = None):
    """Check if the given or stored SENTINEL username and password is 
    correct. Accounts can be created on https://scihub.copernicus.eu/userguide/SelfRegistration.

    Parameters
    ----------
    user_pw : tuple, optional
        ("", "token") to check, if `None` will try to load the 
        password from the keychain, by default None.

    Returns
    -------
    bool
        True if the password works, otherwise False.
    """
    n_max = 3
    succes = False
    n = 1

    while not succes and n <= n_max:

        if not isinstance(user_pw, type(None)):
            username, password = user_pw
        else:
            username, password = get('SENTINEL')

        try:
            api = SentinelAPI(username, password, 'https://apihub.copernicus.eu/apihub')
            _ = api.count(
                        platformname = 'Sentinel-2',
                        producttype = 'S2MSI2A')
            succes = True
        except:
            error = "wrong token."
            succes = False

        if not succes:
            log.warning(f"Try {n}/{n_max} failed.")
            time.sleep(3)
            
        n += 1

    if not succes:
        log.warning(error)

    return succes

def ecmwf_account(user_pw = None):
    """Check if the given or stored ECMWF key is 
    correct. Accounts can be created on https://cds.climate.copernicus.eu/#!/home.

    Parameters
    ----------
    user_pw : tuple, optional
        ("", "key") to check, if `None` will try to load the 
        password from the keychain, by default None.

    Returns
    -------
    bool
        True if the password works, otherwise False.
    """
    n_max = 1
    succes = False
    n = 1

    while not succes and n <= n_max:

        if not isinstance(user_pw, type(None)):
            url, key = user_pw
        else:
            url, key = get('ECMWF')

        try:
            c = cdsapi.Client(url = url, key = key, verify = True, quiet = True)
            fp = os.path.join(pywapor.__path__[0], "test.zip")
            _ = c.retrieve(
                'sis-agrometeorological-indicators',
                {
                    'format': 'zip',
                    'variable': '2m_temperature',
                    'statistic': '24_hour_mean',
                    'year': '2005',
                    'month': '11',
                    'day': ['01'],
                    'area': [2,-2,-2,2]
                },
                fp)
            os.remove(fp)
            succes = True
        except:
            error = "wrong key."
            succes = False

        if not succes:
            log.warning(f"Try {n}/{n_max} failed.")
            time.sleep(3)
            
        n += 1

    if not succes:
        log.warning(error)

    return succes

if __name__ == "__main__":
    nasa_succes = nasa_account()
    vito_succes = vito_account()
    wapor_succes = wapor_account()
    ecmwf_succes = ecmwf_account()
    sentinel_succes = sentinel_account()
    viirsl1_succes = viirsl1_account()

    print(nasa_succes, vito_succes, wapor_succes, ecmwf_succes, 
          viirsl1_succes, sentinel_succes)


import os
import tqdm
from pywapor.general.processing_functions import save_ds, open_ds, process_ds, remove_ds
from osgeo import gdal
import urllib
from pywapor.enhancers.apply_enhancers import apply_enhancer
from pywapor.general.logger import log

def download(fp, product_name, coords, variables, post_processors, url_func, 
                gdal_config_options = {}, waitbar = True, ndv = -9999):

    folder, fn = os.path.split(fp)

    # Create folder.
    if not os.path.isdir(folder):
        os.makedirs(folder)

    for k, v in gdal_config_options.items():
        gdal.SetConfigOption(k, v)

    bands = [int(k.replace("Band", "")) for k in variables.keys() if "Band" in k]

    # Define bounding-box.
    bb = [coords["x"][1][0], coords["y"][1][1], 
            coords["x"][1][1], coords["y"][1][0]]

    options_dict = {
            "projWin": bb,
            "format": "netCDF",
            "creationOptions": ["COMPRESS=DEFLATE", "FORMAT=NC4"],
            "noData": ndv,
            "bandList": bands,
    }

    if waitbar:
        # Create waitbar.
        waitbar = tqdm.tqdm(position = 0, total = 100, bar_format='{l_bar}{bar}|', delay = 20)

        # Define callback function for waitbar progress.
        def _callback_func(info, *args):
            waitbar.update(info * 100 - waitbar.n)

        # Add callback to gdal.Translate options.
        options_dict.update({"callback": _callback_func})

    # Set gdal.Translate options.
    options = gdal.TranslateOptions(**options_dict)

    # Check if url is local or online path.
    url = url_func(product_name)
    is_not_local = urllib.parse.urlparse(url).scheme in ('http', 'https',)
    if is_not_local:
        url = f"/vsicurl/{url}"

    # Run gdal.Translate.
    ds = gdal.Translate(fp.replace(".nc", "_temp.nc"), url, options = options)

    # Reset the gdal.Dataset.
    ds.FlushCache()
    ds = None

    # Process the new netCDF.
    ds = open_ds(fp.replace(".nc", "_temp.nc"))

    ds = ds.rename_vars({k: f"Band{v}" for k,v in zip(list(ds.data_vars), bands)})

    ds = process_ds(ds, coords, variables)

    # Apply product specific functions.
    for var, funcs in post_processors.items():
        for func in funcs:
            ds, label = apply_enhancer(ds, var, func)
            log.info(label)

    # Save final output.
    out = save_ds(ds, fp, encoding = "initiate", label = f"Saving {fn}.")

    # Remove the temporary file.
    remove_ds(ds)

    return out



import tracemalloc
import datetime
from pywapor.general.logger import log, adjust_logger
import types
import numpy as np
import xarray as xr

def format_bytes(size):
    power = 2**10
    n = 0
    power_labels = {0: '', 1: 'K', 2: 'M', 3: 'G', 4: 'T'}
    while size > power:
        size /= power
        n += 1
    return size, power_labels[n]+'B'

def performance_check(func):
    def wrapper_func(*args, **kwargs):
        if "label" in kwargs.keys():
            label = kwargs.pop("label")
        else:
            label = f"`{func.__module__}.{func.__name__}`"
        log.info(f"--> {label}").add()
        t1 = datetime.datetime.now()
        tracemalloc.start()
        out = func(*args, **kwargs)
        mem_test = tracemalloc.get_traced_memory()
        tracemalloc.stop()
        t2 = datetime.datetime.now()
        size, size_label = format_bytes(mem_test[1]-mem_test[0])
        log.info(f"> peak-memory-usage: {size:.1f}{size_label}, execution-time: {t2-t1}.")
        if isinstance(out, xr.Dataset):
            log.info("> chunksize|dimsize: [" + ", ".join([f"{k}: {v[0]}|{sum(v)}" for k, v in out.unify_chunks().chunksizes.items()]) + "]")
        log.sub()
        return out
    wrapper_func.__module__ = func.__module__
    wrapper_func.__name__ = func.__name__
    setattr(wrapper_func, "decorated", True)
    return wrapper_func

def decorate_function(obj, decorator):
    module = obj.__module__
    name = obj.__name__
    if isinstance(obj, types.FunctionType) and not hasattr(obj, 'decorated'):
        setattr(module, name, decorator(obj))

@performance_check
def test(n, k = 100):
    x = np.random.random((n,k,1000))**2
    return x



from functools import partial

def apply_enhancer(ds, variable, enhancer):
    """_summary_

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to enhance.
    variable : str
        Variable name to enhance.
    enhancer : function
        Function to be applied to `variable` inside `ds`. Should take `ds` as
        first argument and `variable` as second.
    source : str, optional
        Specifies the source of the data inside `ds` for logging purposes, by default None.
    log_it : bool, optional
        Create a label to log or not, by default True.

    Returns
    -------
    xr.Dataset
        The enhanced dataset.
    str
        The label to log when calculating the dataset.
    """
    ds = enhancer(ds, variable)

    if isinstance(enhancer, partial):
        func_name = enhancer.func.__name__
    else:
        func_name = enhancer.__name__
    
    if isinstance(variable, type(None)):
        label = f"--> Applying '{func_name}'."
    else:
        label = f"--> Applying '{func_name}' to `{variable}`."

    return ds, label



import os
from pywapor.collect import accounts
import datetime
import pandas as pd
import pywapor.collect.accounts as accounts
from pywapor.general.logger import log
import re
import warnings
from functools import partial
import requests
from pywapor.collect.protocol.crawler import download_urls, crawl
import pywapor.general.bitmasks as bm
import xarray as xr
from pywapor.enhancers.apply_enhancers import apply_enhancer
import numpy as np
from pywapor.general.processing_functions import process_ds, save_ds, open_ds, remove_ds
import datetime
import rioxarray.merge

def download(folder, latlim, lonlim, timelim, product_name, req_vars = ["ndvi", "r0"],
                variables = None, post_processors = None, timedelta = None):

    folder = os.path.join(folder, "PROBAV")

    appending = False
    fn = os.path.join(folder, f"{product_name}.nc")
    if os.path.isfile(fn):
        os.rename(fn, fn.replace(".nc", "_to_be_appended.nc"))
        existing_ds = open_ds(fn.replace(".nc", "_to_be_appended.nc"))
        if np.all([x in existing_ds.data_vars for x in req_vars]):
            existing_ds = existing_ds.close()
            os.rename(fn.replace(".nc", "_to_be_appended.nc"), fn)
            existing_ds = open_ds(fn)
            return existing_ds[req_vars]
        else:
            appending = True
            fn = os.path.join(folder, f"{product_name}_appendix.nc")
            req_vars = [x for x in req_vars if x not in existing_ds.data_vars]

    dates = pd.date_range(timelim[0], timelim[1], freq="D")

    bb = (lonlim[0], latlim[0], lonlim[1], latlim[1])

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    coords = {"x": ["lon", None], "y": ["lat", None]}

    base_url = f"https://www.vito-eodata.be/PDF/datapool/Free_Data/PROBA-V_100m/{product_name}"
    coord_req = "?coord=" + ",".join([str(x) for x in bb])
    url = os.path.join(base_url, coord_req)

    session = requests.sessions.Session()
    session.auth = accounts.get("VITO")

    # Scrape urls.
    urls = find_tiles(url, dates, session)

    # Download .HDF5 tiles.
    fps = download_urls(urls, folder, session, parallel = 4)

    # Convert to netcdf.
    dss = dict()
    for fp in fps:
        date = datetime.datetime.strptime(fp.split("_")[-3], "%Y%m%d")
        ds = open_hdf5_groups(fp, variables, coords).expand_dims({"time": [date]})
        if date in dss.keys():
            dss[date].append(ds)
        else:
            dss[date] = [ds]

    # Merge tiles.
    dss0 = list()
    for date, datasets in dss.items():
        bb = (lonlim[0], latlim[0], lonlim[1], latlim[1])
        ds = rioxarray.merge.merge_datasets(datasets)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category = FutureWarning)
            ds = ds.rio.clip_box(*bb) # NOTE using this seperately, because `bounds`-kw for `merge_datasets` bugs.
        dss0.append(ds)

    # Merge dates.
    ds = xr.concat(dss0, dim = "time")

    # Apply product specific functions.
    for var, funcs in post_processors.items():
        for func in funcs:
            ds, _ = apply_enhancer(ds, var, func)

    if product_name == "S5_TOC_100_m_C1":
        ds["time"] = ds["time"] + np.timedelta64(int(2.5 * 24), "h")

    ds = ds[req_vars]

    ds_new = save_ds(ds, fn, label = f"Merging files.")

    if appending:
        ds = xr.merge([ds_new, existing_ds])
        lbl = f"Appending new variables (`{'`, `'.join(req_vars)}`) to existing file."
        ds = save_ds(ds, os.path.join(folder, f"{product_name}.nc"), encoding = "initiate", label = lbl)
        remove_ds(ds_new)
        remove_ds(existing_ds)
    else:
        ds = ds_new

    for k, v in dss.items():
        for nc in v:
            remove_ds(nc)

    return ds

def open_hdf5_groups(fp, variables, coords):

    nc_fp = fp.replace(".HDF5", ".nc")
    if os.path.isfile(nc_fp):
        ds = open_ds(nc_fp, "all")
    else:
        ds = xr.open_dataset(fp, chunks = "auto")

        spatial_ref_name = [k for k, v in variables.items() if v[1] == "spatial_ref"][0]
        if (spatial_ref_name in list(ds.variables)) and (spatial_ref_name not in list(ds.coords)):
            ds = ds.rio.write_grid_mapping(spatial_ref_name)
            ds = ds.set_coords((spatial_ref_name))

        for k in variables.keys():
            if k in ds.variables:
                continue
            with xr.open_dataset(fp, group = k, engine = "netcdf4", chunks = "auto") as ds_grp:
                vrs = list(ds_grp.variables)[0] # NOTE assuming each group has only 1 variable!
                ds[k] = ds_grp.rename({vrs: k})[k]

        ds = process_ds(ds, coords, variables)

        ds = save_ds(ds, nc_fp, label = f"Converting {os.path.split(fp)[-1]} to netcdf.")

    return ds

def mask_bitwise_qa(ds, var, flags):
    flag_bits = bm.PROBAV_qa_translator()
    mask = bm.get_mask(ds["qa"].astype("uint8"), flags, flag_bits)
    ds[var] = ds[var].where(~mask, np.nan)
    return ds

def calc_r0(ds, *args):
    ds["r0"] = 0.429 * ds["blue"] + 0.333 * ds["red"] + 0.133 * ds["nir"] + 0.105 * ds["swir"]
    return ds

def default_post_processors(product_name, req_vars = ["ndvi", "r0"]):

    post_processors = {
        "S5_TOC_100_m_C1": {
            "r0": [
                    calc_r0, 
                    partial(mask_bitwise_qa, 
                    flags = ["bad BLUE", "bad RED", "bad NIR", "bad SWIR", 
                            "sea", "undefined", "cloud", "ice/snow", "shadow"]),
            ],
            "ndvi": [
                    partial(mask_bitwise_qa, 
                    flags = ["bad RED", "bad NIR", "sea", "undefined", 
                    "cloud", "ice/snow", "shadow"]),
            ],
        }
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def default_vars(product_name, req_vars = ["ndvi", "r0"]):
    
    variables = {
        "S5_TOC_100_m_C1": {
            "LEVEL3/NDVI":              [("lon", "lat"), "ndvi"],
            "LEVEL3/RADIOMETRY/BLUE":   [("lon", "lat"), "blue"],
            "LEVEL3/RADIOMETRY/NIR":    [("lon", "lat"), "nir"],
            "LEVEL3/RADIOMETRY/RED":    [("lon", "lat"), "red"],
            "LEVEL3/RADIOMETRY/SWIR":   [("lon", "lat"), "swir"],
            "LEVEL3/GEOMETRY/VNIR":     [("lon", "lat"), "vnir_vza"],
            "LEVEL3/GEOMETRY/SWIR":     [("lon", "lat"), "swir_vza"],
            "LEVEL3/QUALITY":           [("lon", "lat"), "qa"],
            "crs":                      [(), "spatial_ref"]
        },
    }

    req_dl_vars = {
        "S5_TOC_100_m_C1": {
            "ndvi": ["LEVEL3/NDVI", "LEVEL3/QUALITY", "crs"],
            "r0":   ["LEVEL3/RADIOMETRY/BLUE", "LEVEL3/RADIOMETRY/NIR", 
                    "LEVEL3/RADIOMETRY/RED", "LEVEL3/RADIOMETRY/SWIR",
                    "LEVEL3/QUALITY", "crs"],
        }
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def find_tiles(url, dates, session):

    log.info("--> Searching PROBAV tiles.")

    regex = "https:.*\/\d{4}\/\?coord="
    filter_regex = "\d{4}(?=\/\?coord=)"
    urls = {"_": url}
    label_filter = dates.strftime("%Y")
    years = crawl(urls, regex, filter_regex, session, label_filter = label_filter)

    regex = "https:.*\/\d{4}\/\d{2}\/\?coord="
    filter_regex = "\d{4}\/\d{2}(?=\/\?coord=)"
    label_filter = dates.strftime("%Y%m")
    months = crawl(years, regex, filter_regex, session, label_filter = label_filter)

    regex = "https:.*\/\d{4}\/\d{2}\/\d{2}\/\?coord="
    filter_regex = "\d{4}\/\d{2}\/\d{2}(?=\/\?coord=)"
    label_filter = dates.strftime("%Y%m%d")
    days = crawl(months, regex, filter_regex, session, label_filter = label_filter)

    regex = "https:.*\/\d{4}\/\d{2}\/\d{2}\/.*\/\?coord="
    filter_regex = "\d{4}\/\d{2}\/\d{2}(?=\/.*\/\?coord=)"
    label_filter = dates.strftime("%Y%m%d")
    prods = crawl(days, regex, filter_regex, session, label_filter = label_filter)

    regex = ".*\.HDF5"
    filter_regex = "\d{8}(?=.*\.HDF5)"
    fns = crawl(prods, regex, filter_regex, session, list_out = True)

    dl_urls = [os.path.join(re.sub("\?coord=.*","",prods[date_str]), fn) for date_str, fn, in fns]

    log.info(f"--> Downloading {len(dl_urls)} PROBAV tiles.")

    return dl_urls

if __name__ == "__main__":

    product_name = "S5_TOC_100_m_C1"

    folder = r"/Users/hmcoerver/Downloads/pywapor_test"
    # latlim = [26.9, 33.7]
    # lonlim = [25.2, 37.2]
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    timelim = [datetime.date(2020, 7, 2), datetime.date(2020, 7, 9)]

    variables = None
    post_processors = None
    req_vars = ["r0", "ndvi"]

    # ds = download(folder, latlim, lonlim, timelim, product_name, req_vars = req_vars,
    #             variables = variables, post_processors = post_processors)




"""
Functions to create an interactive network graph based on metadata
attributes in the xr.Dataset output by `pywapor.et_look`.
"""

from pyvis.network import Network
import json

def make_node(ds, var):
    """Takes available metadata from a `var` in `ds` and formats it for use
    with pyvis.network.Network.add_node().

    Parameters
    ----------
    ds : xr.Dataset
        Dataset from which to take `var`.
    var : str
        Variable for which to prepare the network node.

    Returns
    -------
    dict
        Information required to create a network node.
    """

    if var in list(ds.variables):
        if "units" in ds[var].attrs.keys():
            units = ds[var].units
        else:
            units = ""

        if "long_name" in ds[var].attrs.keys():
            long_name = ds[var].long_name
        else:
            long_name = ""

        if "et_look_module" in ds[var].attrs.keys():
            group = ds[var].et_look_module
            shape = "dot"
            module = ds[var].et_look_module
        else:
            if len(ds[var].coords) == 0:
                group = "scalar"
            elif len(ds[var].coords) == 2:
                group = "temporal-constant"
            elif len(ds[var].coords) == 3:
                group = "temporal-input"
            elif len(ds[var].coords) > 3:
                group = "main-input"
            else:
                group = "unknown"
            shape = "square"
            module = "pre_et_look"

        if "calculated_with" in ds[var].attrs.keys():

            upstreams = ds[var].calculated_with
            if not isinstance(upstreams, list):
                upstreams = [upstreams]

            f = " = f(" + ", ".join(upstreams) + ")"
        else:
            f = ""


    else:
        long_name = ""
        units = ""
        module = ""
        group = ""
        module = ""
        f = ""
        shape = "dot"
        
    title = f"""
    <u>{long_name}</u><br><br>
    <b>{var}</b>{f} <br><br>
    Units: [{units}] <br>
    Module: {module}
    """

    return {"n_id": var, "label": var, "title": title, "group": group, "shape": shape}

def create_network(ds, fh, exclude = ["scalar", "temporal-input"]):
    """Creates a network graph for a dataset output by et_look.

    Parameters
    ----------
    ds : xr.Dataset
        Output of `et_look`.
    fh : str
        Path to where to save the network graph. Extension should be `.html`.
    exclude : list, optional
        Which types of variables to exclude from the graph. `scalar` excludes
        variables that have neither a spatial nor temporal dimension. `temporal-input`
        excludes variables that have only a temporal dimension. `temporal-constant`
        excludes variables that have only spatial dimensions. By default ["scalar", "temporal-input"].
    """
    net = Network(width = "100%", height = "100%", layout = False, 
                    bgcolor='#E9ECF5', font_color='black', directed = True)

    options = {
        "physics": {
            "hierarchicalRepulsion": {
                "centralGravity": 0,
            },
        "minVelocity": 0.75,
        "solver": "hierarchicalRepulsion",
        },
        "edges": {
            "color": {
                "color": "#B5BFDE",
                "highlight": "black",
            },
            "selectionWidth": 4,
            "smooth": {
                "type": "cubicBezier",
            }
        },
        "layout": {
            "randomSeed": 460885,
            "improvedLayout": True,
        },
        "groups": network_colors(),
        "interaction": {
            "hover": True,
            "tooltipDelay": 10,
            # "zoomSpeed": 3, doesnt work.
            "navigationButtons": False,
        }
    }

    net.set_options(json.dumps(options))

    for var in list(ds.variables):

        if "calculated_with" in ds[var].attrs.keys():

            upstreams = ds[var].calculated_with
            if not isinstance(upstreams, list):
                upstreams = [upstreams]

            if not var in net.get_nodes():
                kwargs = make_node(ds, var)
                if kwargs["group"] in exclude:
                    continue
                net.add_node(**kwargs)

            for upstream in upstreams:
                if not upstream in net.get_nodes():
                    kwargs = make_node(ds, upstream)
                    if kwargs["group"] in exclude:
                        continue
                    net.add_node(**kwargs)

                net.add_edge(upstream, var)

    net.show(fh)

def network_colors():
    group_configs = {
            "main-input": {
                "color": {
                "background": "#122332",
                "border": "#14293B",
                "highlight": {
                    "border": "#2A4257",
                    "background": "#183751"
                },
                "hover": {
                    "background": "#2A4257",
                    "border": "#183751"
                }
                },
                "borderWidth": 1,
                "borderWidthSelected": 2
            },
            "pywapor.et_look_v2.meteo": {
                "color": {
                "border": "#2B7CE9",
                "background": "#97C2FC",
                "highlight": {
                    "border": "#2B7CE9",
                    "background": "#D2E5FF"
                },
                "hover": {
                    "border": "#2B7CE9",
                    "background": "#D2E5FF"
                }
                }
            },
            "pywapor.et_look_v2.radiation": {
                "color": {
                "border": "#FFA500",
                "background": "#FFFF00",
                "highlight": {
                    "border": "#FFA500",
                    "background": "#FFFFA3"
                },
                "hover": {
                    "border": "#FFA500",
                    "background": "#FFFFA3"
                }
                }
            },
            "pywapor.et_look_v2.solar_radiation": {
                "color": {
                "border": "#FA0A10",
                "background": "#FB7E81",
                "highlight": {
                    "border": "#FA0A10",
                    "background": "#FFAFB1"
                },
                "hover": {
                    "border": "#FA0A10",
                    "background": "#FFAFB1"
                }
                }
            },
            "pywapor.et_look_v2.roughness": {
                "color": {
                "border": "#41A906",
                "background": "#7BE141",
                "highlight": {
                    "border": "#41A906",
                    "background": "#A1EC76"
                },
                "hover": {
                    "border": "#41A906",
                    "background": "#A1EC76"
                }
                }
            },
            "pywapor.et_look_v2.leaf": {
                "color": {
                "border": "#E129F0",
                "background": "#EB7DF4",
                "highlight": {
                    "border": "#E129F0",
                    "background": "#F0B3F5"
                },
                "hover": {
                    "border": "#E129F0",
                    "background": "#F0B3F5"
                }
                }
            },
            "pywapor.et_look_v2.unstable": {
                "color": {
                "border": "#7C29F0",
                "background": "#AD85E4",
                "highlight": {
                    "border": "#7C29F0",
                    "background": "#D3BDF0"
                },
                "hover": {
                    "border": "#7C29F0",
                    "background": "#D3BDF0"
                }
                }
            },
            "pywapor.et_look_v2.resistance": {
                "color": {
                "border": "#C37F00",
                "background": "#FFA807",
                "highlight": {
                    "border": "#C37F00",
                    "background": "#FFCA66"
                },
                "hover": {
                    "border": "#C37F00",
                    "background": "#FFCA66"
                }
                }
            },
            "pywapor.et_look_v2.neutral": {
                "color": {
                "border": "#4220FB",
                "background": "#6E6EFD",
                "highlight": {
                    "border": "#4220FB",
                    "background": "#9B9BFD"
                },
                "hover": {
                    "border": "#4220FB",
                    "background": "#9B9BFD"
                }
                }
            },
            "pywapor.et_look_v2.evapotranspiration": {
                "color": {
                "border": "#FD5A77",
                "background": "#FFC0CB",
                "highlight": {
                    "border": "#FD5A77",
                    "background": "#FFD1D9"
                },
                "hover": {
                    "border": "#FD5A77",
                    "background": "#FFD1D9"
                }
                }
            },
            "temporal-constant": {
                "color": {
                "border": "#4AD63A",
                "background": "#C2FABC",
                "highlight": {
                    "border": "#4AD63A",
                    "background": "#E6FFE3"
                },
                "hover": {
                    "border": "#4AD63A",
                    "background": "#E6FFE3"
                }
                }
            },
            "pywapor.et_look_dev.biomass": {
                "color": {
                "border": "#990000",
                "background": "#EE0000",
                "highlight": {
                    "border": "#BB0000",
                    "background": "#FF3333"
                },
                "hover": {
                    "border": "#BB0000",
                    "background": "#FF3333"
                }
                }
            },
            "pywapor.et_look_dev.leaf": {
                "color": {
                "border": "#FF6000",
                "background": "#FF6000",
                "highlight": {
                    "border": "#FF6000",
                    "background": "#FF6000"
                },
                "hover": {
                    "border": "#FF6000",
                    "background": "#FF6000"
                }
                }
            },
            "pywapor.et_look_v2.stress": {
                "color": {
                "border": "#97C2FC",
                "background": "#2B7CE9",
                "highlight": {
                    "border": "#D2E5FF",
                    "background": "#2B7CE9"
                },
                "hover": {
                    "border": "#D2E5FF",
                    "background": "#2B7CE9"
                }
                }
            }
            }
    return group_configs

if __name__ == "__main__":

    import xarray as xr

    fh = r"/Users/hmcoerver/Downloads/pywapor_v2_network.html"

    ds = xr.open_dataset(r"/Users/hmcoerver/pywapor_notebooks/level_1/et_look_output___.nc")

    create_network(ds, fh)

    print("done")


import copy
import pywapor.se_root as se_root
from functools import partial
import types
import numpy as np
from pywapor.general.logger import log
from pywapor.enhancers.dms.thermal_sharpener import sharpen

def find_setting(sources, setting_str, max_length = np.inf, min_length = 0):
    example_sources = list()
    for var, x in sources.items():
        prod = [product for product in x["products"] if setting_str in product.keys()]
        if len(prod) >= 1:
            for pro in prod:
                if pro[setting_str]:
                    example_source = (pro["source"], pro["product_name"])
                    if isinstance(example_source[0], types.FunctionType):
                        example_source = (example_source[0].__name__, example_source[1])
                    elif isinstance(example_source[0], partial):
                        example_source = (example_source[0].func.__name__, example_source[1])
                    if example_source not in example_sources:
                        example_sources.append(example_source)
        else:
            if setting_str in x.keys():
                example_sources.append(var)
    if len(example_sources) > max_length:
            log.warning(f"--> Found more than {max_length} products for `{setting_str}`.")
    if len(example_sources) > max_length or len(example_sources) < min_length:
        log.warning(f"--> Didn't find any products for `{setting_str}`.")
    return example_sources

def pre_et_look_levels(level = "level_1", bin_length = "DEKAD"):

    se_root_dler = partial(se_root.se_root, bin_length = bin_length, 
                            sources = level)

    level_1 = {

        "ndvi": {
            "products": [
                {
                    "source": "MODIS",
                    "product_name": "MOD13Q1.061",
                    "enhancers": "default",
                    "is_example": True
                },
                {
                    "source": "MODIS",
                    "product_name": "MYD13Q1.061",
                    "enhancers": "default",
                }
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "nearest",
            },

        "r0": {
            "products": [
                {
                    "source": "MODIS",
                    "product_name": "MCD43A3.061",
                    "enhancers": "default",
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "se_root": {
            "products": [
                {
                    "source": se_root_dler,
                    "product_name": "v2",
                    "enhancers": "default",
                },
            ],
            "composite_type": "max",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
        },

        "p": {
            "products": [
                {
                    "source": "CHIRPS",
                    "product_name": "P05",
                    "enhancers": "default",
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "z": {
            "products": [
                {
                    "source": "SRTM",
                    "product_name": "30M",
                    "enhancers": "default",
                },
            ],
            "composite_type": None,
            "temporal_interp": None,
            "spatial_interp": "bilinear",
            },

        "ra": {
            "products": [
                {
                    "source": "MERRA2",
                    "product_name": "M2T1NXRAD.5.12.4",
                    "enhancers": "default",
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "t_air": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "t_air_max": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "composite_type": "max",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "t_air_min": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "composite_type": "min",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "u2m": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "v2m": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "qv": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "p_air": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "p_air_0": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "wv": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "land_mask": {
            "products": [
                {
                    "source": "GLOBCOVER",
                    "product_name": "2009_V2.3_Global",
                    "enhancers": "default",
                },
            ],
            "composite_type": None,
            "temporal_interp": None,
            "spatial_interp": "nearest",
            },

        "rs_min": {
            "products": [
                {
                    "source": "GLOBCOVER",
                    "product_name": "2009_V2.3_Global",
                    "enhancers": "default",
                },
            ],
            "composite_type": None,
            "temporal_interp": None,
            "spatial_interp": "nearest",
            },

        "z_obst_max": {
            "products": [
                {
                    "source": "GLOBCOVER",
                    "product_name": "2009_V2.3_Global",
                    "enhancers": "default",
                },
            ],
            "composite_type": None,
            "temporal_interp": None,
            "spatial_interp": "nearest",
            },

    }

    statics = [
                'lw_offset', 'lw_slope', 'z_oro', 'rn_offset', 'rn_slope', 't_amp_year', 't_opt', 'vpd_slope',
                # 'land_mask', 'rs_min', 'z_obst_max' # NOTE generated from lulc
            ]

    for var in statics:

        level_1[var] =  {
            "products": [
                {
                    "source": "STATICS",
                    "product_name": "WaPOR2",
                    "enhancers": "default",
                },
            ],
            "composite_type": None,
            "temporal_interp": None,
            "spatial_interp": "bilinear",
        }

    level_2 = copy.deepcopy(level_1)

    level_2["ndvi"] = {

            "products": [
                {
                    "source": "PROBAV",
                    "product_name": "S5_TOC_100_m_C1",
                    "enhancers": "default",
                    "is_example": True
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "nearest",
            }

    level_2["r0"] = {

            "products": [
                {
                    "source": "PROBAV",
                    "product_name": "S5_TOC_100_m_C1",
                    "enhancers": "default",
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "nearest",
            }

    level_2_v3 = dict()

    level_2_v3["ndvi"] = {

            "products": [
                {
                    "source": "SENTINEL2",
                    "product_name": "S2MSI2A_R60m",
                    "enhancers": "default",
                    "is_example": True
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "nearest",
            }

    level_2_v3["r0"] = {

            "products": [
                {
                    "source": "SENTINEL2",
                    "product_name": "S2MSI2A_R60m",
                    "enhancers": "default",
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "nearest",
            }

    level_2_v3["se_root"] = {
            "products": [
                {
                    "source": se_root_dler,
                    "product_name": "v3",
                    "enhancers": "default",
                },
            ],
            "composite_type": "max",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
        }

    level_2_v3["p"] = {
        "products": [
            {
                "source": "CHIRPS",
                "product_name": "P05",
                "enhancers": "default",
            },
        ],
        "composite_type": "mean",
        "temporal_interp": "linear",
        "spatial_interp": "bilinear",
        }

    level_2_v3["z"] = {
        "products": [
            {
                "source": "COPERNICUS",
                "product_name": "GLO90",
                "enhancers": "default",
            },
        ],
        "composite_type": None,
        "temporal_interp": None,
        "spatial_interp": "bilinear",
        }

    for var, composite_type in [("t_air", "mean"), ("t_air_min", "min"), ("t_air_max", "max"), 
                                # ("t_dew", "mean"), # NOTE ETLook usees t_dew to calc `vp`, but that one is directly available from agERA5.
                                # ("rh", "mean"), 
                                ("u", "mean"), ("vp", "mean"), ("ra", "mean")]:
        level_2_v3[var] = {
            "products": [
                {
                    "source": "ERA5",
                    "product_name": "sis-agrometeorological-indicators",
                    "enhancers": "default",
                },
            ],
            "composite_type": composite_type,
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            }

    for var in ["p_air", "p_air_0"]:
        level_2_v3[var] = {
            "products": [
                {
                    "source": "ERA5",
                    "product_name": 'reanalysis-era5-single-levels',
                    "enhancers": "default",
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            }

    statics_v2 = [
                'z_oro', 'rn_offset', 'rn_slope', 't_amp_year', 't_opt', 'vpd_slope',
                'land_mask', 'rs_min', 'z_obst_max'
            ]

    for var in statics_v2:
        level_2_v3[var] = {
            'products': [
                {
                    'source': "STATICS",
                    'product_name': "WaPOR2",
                    'enhancers': 'default'
                },
            ],
        'composite_type': None,
        'temporal_interp': None,
        'spatial_interp': 'bilinear'}

    statics_v3 = [
                'lw_offset', 'lw_slope'
            ]

    for var in statics_v3:
        level_2_v3[var] = {
            'products': [
                {
                    'source': "STATICS",
                    'product_name': "WaPOR3",
                    'enhancers': 'default'
                },
            ],
        'composite_type': None,
        'temporal_interp': None,
        'spatial_interp': 'bilinear'}
    



    levels = {
            "level_1": level_1,
            "level_2": level_2,
            "level_2_v3": level_2_v3
                }

    return levels[level]

def pre_se_root_levels(level = "level_1"):

    level_1 = {

        "ndvi": {
            "products": [
                {
                    "source": "MODIS",
                    "product_name": "MOD13Q1.061",
                    "enhancers": "default",
                    "is_example": True,
                },
                {
                    "source": "MODIS",
                    "product_name": "MYD13Q1.061",
                    "enhancers": "default", 
                }
            ],
            "temporal_interp": "linear",
            "spatial_interp": "nearest",
            },

        "lst": {
            "products": [
                {
                    "source": "MODIS",
                    "product_name": "MOD11A1.061",
                    "enhancers": "default",
                },
                {
                    "source": "MODIS",
                    "product_name": "MYD11A1.061",
                    "enhancers": "default",
                },
            ],
            "temporal_interp": None,
            "spatial_interp": "nearest",
        },

        "t_air": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "t_air_max": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "t_air_min": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "u2m": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "v2m": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "qv": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "wv": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "p_air": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },

        "p_air_0": {
            "products": [
                {
                    "source": "GEOS5",
                    "product_name": "inst3_2d_asm_Nx",
                    "enhancers": "default",
                },
            ],
            "temporal_interp": "linear",
            "spatial_interp": "bilinear",
            },
    }

    statics = ["r0_bare", "r0_full"]

    for var in statics:

        level_1[var] =  {
            "products": [
                {
                    "source": "STATICS",
                    "product_name": "WaPOR2",
                    "enhancers": "default",
                },
            ],
            "temporal_interp": None,
            "spatial_interp": "bilinear",
        }

    level_2 = copy.deepcopy(level_1)

    level_2["ndvi"] = {

            "products": [
                {
                    "source": "PROBAV",
                    "product_name": "S5_TOC_100_m_C1",
                    "enhancers": "default",
                    "is_example": True
                },
            ],
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "nearest",
            }

    level_2_v3 = dict()

    level_2_v3["ndvi"] = {
        "products": [
            {
                "source": "SENTINEL2",
                "product_name": "S2MSI2A_R60m",
                "enhancers": "default",
                "is_example": True
            },
        ],
        "temporal_interp": "linear",
        "spatial_interp": "nearest"}

    for var in ["mndwi", "psri", "vari_red_edge", "bsi", "nmdi", "green", "nir"]:
        level_2_v3[var] = {
            'products': [{
                'source': 'SENTINEL2',
                'product_name': 'S2MSI2A_R60m',
                'enhancers': 'default',
            }],
        'temporal_interp': 'linear',
        'spatial_interp': 'nearest'}

    level_2_v3['bt'] = {
        'products': [
            {
                'source': 'VIIRSL1',
                'product_name': 'VNP02IMG',
                'enhancers': 'default'
            },
        ],
        'variable_enhancers': [sharpen],
        'temporal_interp': None,
        'spatial_interp': 'nearest'}

    for var in ["u", "t_dew", "p_air_0", "p_air", "t_air", "wv"]:
        level_2_v3[var] = {
            'products': [
                {
                    'source': 'ERA5',
                    'product_name': 'reanalysis-era5-single-levels',
                    'enhancers': 'default'
                },
            ],
        'temporal_interp': "linear",
        'spatial_interp': 'bilinear'}

    for var in ["z", "slope", "aspect"]:
        level_2_v3[var] = {
            'products': [
                {
                    'source': 'COPERNICUS',
                    'product_name': 'GLO90',
                    'enhancers': 'default'
                },
            ],
        'temporal_interp': None,
        'spatial_interp': 'bilinear'}

    for var in ["r0_bare", "r0_full"]:
        level_2_v3[var] = {
            'products': [
                {
                    'source': "STATICS",
                    'product_name': "WaPOR2",
                    'enhancers': 'default'
                },
            ],
        'temporal_interp': "linear",
        'spatial_interp': 'bilinear'}


    levels = {
                "level_1": level_1,
                "level_2": level_2,
                "level_2_v3": level_2_v3,
                }

    return levels[level]

if __name__ == "__main__":

    et_look_sources_lvl1 = pre_et_look_levels(level = "level_1", bin_length = "DEKAD")
    et_look_sources_lvl2 = pre_et_look_levels(level = "level_2", bin_length = "DEKAD")
    et_look_sources_lvl2_v3 = pre_et_look_levels(level = "level_2_v3", bin_length = "DEKAD")

    se_root_sources_lvl1 = pre_se_root_levels(level = "level_1")
    se_root_sources_lvl2 = pre_se_root_levels(level = "level_2")
    se_root_sources_lvl2_v3 = pre_se_root_levels(level = "level_2_v3")

    